from copy import deepcopy
from sklearn.base import BaseEstimator
import pandas as pd
from ..helpers.ML_Helpers import conv_to_list, proc_input, proc_type_dep_str

from ..helpers.VARS import ORDERED_NAMES
from ..main.Input_Tools import (is_duplicate, is_pipe, is_select,
                                is_special, is_value_subset)
from ..pipeline.Scorers import process_scorers


def proc_all(base_obj):

    if base_obj is None:
        return

    proc_name(base_obj, 'obj')
    proc_name(base_obj, 'base_model')
    proc_name(base_obj, 'models')
    proc_name(base_obj, 'scorer')


def proc_name(base_obj, name):

    if hasattr(base_obj, name):
        obj = getattr(base_obj, name)

        if isinstance(obj, list):
            for o in obj:
                proc_all(o)

        elif hasattr(obj, 'obj'):
            proc_all(obj)

        elif isinstance(obj, str):
            setattr(base_obj, name, proc_input(obj))


class Params(BaseEstimator):

    def set_values(self, args):

        for key in args:
            if key != 'self' and hasattr(self, key):

                try:
                    if isinstance(args[key], str):
                        if args[key] != 'default':
                            setattr(self, key, args[key])
                    else:
                        setattr(self, key, args[key])

                # If value error, set to key
                except ValueError:
                    setattr(self, key, args[key])

    def copy(self):
        return deepcopy(self)


class Check():

    def check_args(self):

        if hasattr(self, 'obj'):
            self._check_obj()

        if hasattr(self, 'params'):
            self._check_params()

        if hasattr(self, 'obj') and hasattr(self, 'params'):
            self._check_obj_and_params()

        if hasattr(self, 'scope'):
            self._check_scope()

        if hasattr(self, 'extra_params'):
            self._check_extra_params()

        if hasattr(self, 'base_model'):
            self._check_base_model()

        try:
            self.check_extra_args()
        except AttributeError:
            pass

    def _check_obj(self):

        obj = getattr(self, 'obj')

        if obj is None:
            raise IOError('passed obj cannot be None, to ignore',
                          'the object itself',
                          'set it within Model_Pipeline to None,',
                          ' not obj here.')

        if isinstance(obj, list) and not is_pipe(obj):
            raise IOError('You may only pass a list of objs with ',
                          'special input',
                          'Pipe()')

    def _check_params(self):

        params = getattr(self, 'params')

        if params is None:
            setattr(self, 'params', 0)

    def _check_obj_and_params(self):

        obj = getattr(self, 'obj')
        params = getattr(self, 'params')

        if is_pipe(obj):

            if not isinstance(params, list):

                if params == 0:
                    new_params = [0 for i in range(len(obj))]
                    setattr(self, 'params', new_params)
                else:
                    raise IOError('obj is passed as Pipe, but params was not.',
                                  'Make sure',
                                  'params is either a list or Pipe ',
                                  'with the same length as',
                                  'self.obj')

            elif len(params) != len(obj):

                if params == [0]:
                    new_params = [0 for i in range(len(obj))]
                    setattr(self, 'params', new_params)
                else:
                    raise IOError('obj is passed as Pipe, Make sure',
                                  'params is either a list or Pipe with the ',
                                  'same length as',
                                  'self.obj')

    def _check_scope(self):

        pass
        # Nothing to check as of right now
        # scope = getattr(self, 'scope')

    def _check_extra_params(self):

        extra_params = getattr(self, 'extra_params')
        if extra_params is None:
            return
        #    setattr(self, 'extra_params', {})

        elif not isinstance(extra_params, dict):
            raise IOError('extra params must be a dict!')

    def _check_base_model(self):

        base_model = getattr(self, 'base_model')

        # None is okay
        if base_model is None:
            return

        # Assume select is okay too
        if is_select(base_model):
            return

        if not hasattr(base_model, '_is_model'):
            raise IOError('base_model must be either None or a valid '
                          'Model / Ensemble ',
                          'set of wrapeper params!')


class Piece(Params, Check):
    pass


class Loader(Piece):

    def __init__(self, obj, params=0, scope='data files',
                 cache_loc=None, extra_params=None,
                 fix_n_wrapper_jobs=False):
        ''' Loader refers to transformations which operate on loaded Data_Files.
        (See :func:`Load_Data_Files`).
        They in essence take in saved file locations, and after some series
        of transformations pass on compatible features.
        Notably loaders
        define operations which are computed on single files indepedently.

        Parameters
        ----------
        obj : str, custom obj or :class:`Pipe`
            `obj` selects the base loader object to use, this can be either
            a str corresponding to
            one of the preset loaders found at :ref:`Loaders`.
            Beyond pre-defined loaders, users can
            pass in custom objects
            (they just need to have a defined fit_transform function
            which when passed the already loaded file, will return
            a 1D representation of that subjects
            features.

            `obj` can also be passed as a :class:`Pipe`.
            See :class:`Pipe`'s documentation to
            learn more on how this works, and why you might want to use it.

            See :ref:`Pipeline Objects` to read more
            about pipeline objects in general.

            For example, the 'identity' loader will load in saved data at
            the stored file
            location, lets say they are 2d numpy arrays,
            and will return a flattened version
            of the saved arrays, with each data point as a feature.
            A more practical example
            might constitute loading in say 3D neuroimaging data,
            and passing on features as extracted by ROI.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` determines optionally if the distribution
            of hyper-parameters to
            potentially search over for this loader. Preset param distributions are
            listed for each choice of obj at :ref:`Loaders`, and you can read more on
            how params work more generally at :ref:`Params`.

            If obj is passed as :class:`Pipe`, see :class:`Pipe` for an example on how different
            corresponding params can be passed to each piece individually.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified loader
            should transform. See :ref:`Scopes` for more information on
            how scopes can
            be specified.

            You will likely want to use either custom key based scopes, or the
            'data files' preset scope, as something like 'covars'
            won't make much sense,
            when atleast for now, you cannot even load Covars data files.

            ::

                default = 'data files'

        cache_loc : str, Path or None, optional
            Optional location in which to cache loader transformations.

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        fix_n_wrapper_jobs : int or False, optional
            Typically this parameter is left as default, but
            in special cases you may want to set this. It controls
            the number of jobs fixed for the Loading Wrapper.

            This parameter can be used to set that value.

            ::

                default = False
        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.extra_params = extra_params
        self.fix_n_wrapper_jobs = fix_n_wrapper_jobs

        self.check_args()


class Imputer(Piece):

    def __init__(self, obj, params=0, scope='all',
                 base_model=None, base_model_type='default',
                 extra_params=None):
        ''' If there is any missing data (NaN's) that have been kept
        within data or covars, then an imputation strategy must be
        defined! This object allows you to define an imputation strategy.
        In general, you should need at most two Imputers, one for all
        `float` type data and one for all categorical data, assuming you
        have been present, and they both have missing values.

        Parameters
        ----------
        obj : str
            `obj` selects the base imputation strategy to use.
            See :ref:`Imputers` for all avaliable options.
            Notably, if 'iterative' is passed,
            then a base model must also be passed!
            Also note that the `sample_posterior`
            argument within `iterative` imputer is not currently supported.

            See :ref:`Pipeline Objects` to read more about
            pipeline objects in general.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` set an associated distribution of hyper-parameters to
            potentially search over with the Imputer.
            Preset param distributions are
            listed for each choice of params with the corresponding
            obj at :ref:`Imputers`,
            and you can read more on how params
            work more generally at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified
            imputer will have access to.

            The main options that make sense for imputer are
            one for `float` data and one for `categorical` / 'cat' datatypes.
            Though you can also pass a custom set of keys.

            Note: If using iterative imputation you may want to carefully
            consider the scope passed. For example, while it may be beneficial
            to impute categorical and float features seperately, i.e., with
            different base_model_type's
            (categorical for categorical and regression for float), you must
            also consider that in predicting the missing values under
            this setup, the categorical imputer would not have access to
            to the float features and vice versa.

            In this way, you
            may want to either just treat all features as float, or
            instead of imputing categorical features, load missing
            values as a seperate category - and then set the scope
            here to be 'all', such that the iterative imputer has
            access to all features. Essently why this is neccisary
            is the iterative imputer
            will try to replace any NaN value present in its input
            features.

            See :ref:`Scopes` for more information on how scopes can
            be specified.

            ::

                default = 'all'

        scope : {'float', 'cat', custom}, optional
            `scope` determines on which subset of features
            the imputer should act on.

            :ref:`Scopes`.

            ::

                default = 'float'

        base_model : :class:`Model`, :class:`Ensemble` or None, optional
            If 'iterative' is passed to obj, then a base_model is required in
            order to perform iterative imputation! The base model can be
            any valid Model_Pipeline Model.

            ::

                default = None

        base_model_type : 'default' or Problem Type, optional
            In setting a base imputer model, it may be desirable to
            have this model have a different 'problem type', then your
            over-arching problem. For example, if performing iterative
            imputation on categorical features only, you will likely
            want to use a categorical predictor - but for imputing on
            float-type features, you will want to use a 'regression' type
            base model.

            Choices are {'binary', 'regression', 'categorical'} or 'default'.
            If 'default', then the following behavior will be applied:
            If the scope of the imputer is set to 'cat' or 'categorical',
            then the 'categorical' problem type will be used for the base
            model. If anything else, then the 'regression' type will be used.

            ::

                default = 'default'


        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.base_model = deepcopy(base_model)
        self.base_model_type = base_model_type
        self.extra_params = extra_params

        self.check_args()


class Scaler(Piece):

    def __init__(self, obj, params=0, scope='float', extra_params=None):
        ''' Scaler refers to a piece in the :class:`Model_Pipeline`,
        which is responsible
        for performing any sort of scaling or transformation on the data
        which doesn't require the
        target variable, and doesn't change the number of data
        points or features.

        Parameters
        ----------
        obj : str or custom obj
            `obj` if passed a str selects a scaler from the preset defined
            scalers, See :ref:`Scalers`
            for all avaliable options. If passing a custom object, it must
            be a sklearn compatible
            transformer, and further must not require the target variable,
            not change the number of data
            points or features.

            See :ref:`Pipeline Objects` to read more about
            pipeline objects in general.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` set an associated distribution of hyper-parameters to
            potentially search over with this Scaler.
            Preset param distributions are
            listed for each choice of params with the
            corresponding obj at :ref:`Scalers`,
            and you can read more on how params work more
            generally at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified scaler
            should transform. See :ref:`Scopes` for more
            information on how scopes can
            be specified.

            ::

                default = 'float'

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.extra_params = extra_params

        self.check_args()


class Transformer(Piece):

    def __init__(self, obj, params=0, scope='float', cache_loc=None,
                 extra_params=None, fix_n_wrapper_jobs='default'):
        ''' The Transformer is base optional component of the
        :class:`Model_Pipeline` class.
        Transformers define any type of transformation to the loaded
        data which may change the number
        of features in a non-simple way (i.e., conceptually distinct
        from :class:`Feat_Selector`, where you
        know in advance the transformation is just selecting a subset
        of existing features). These are
        transformations like applying Principle Component Analysis,
        or on the fly One Hot Encoding.

        Parameters
        ----------
        obj : str or custom_obj
            `obj` if passed a str selects from the avaliable class
            defined options for
            transformer as found at :ref:`Transformers`.

            If a custom object is passed as `obj`, it must be a
            sklearn api compatible
            transformer (i.e., have fit, transform, get_params
            and set_params methods, and further be
            cloneable via sklearn's clone function).
            See :ref:`Custom Input Objects` for more info.

            See :ref:`Pipeline Objects` to read more about
            pipeline objects in general.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` determines optionally if the distribution of
            hyper-parameters to potentially search over for this transformer.
            Preset param distributions are listed for each choice of obj
            at :ref:`Transformers`, and you can read more on
            how params work more generally at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features
            the specified transformer
            should transform. See :ref:`Scopes` for more information
            on how scopes can
            be specified.

            Specifically, it may be useful to consider the use of
            :class:`Duplicate` here.

            ::

                default = 'float'

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        fix_n_wrapper_jobs : int or 'default', optional
            This parameter is ignored right now for Transformers

            ::

                default = 'default'
        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.extra_params = extra_params
        self.fix_n_wrapper_jobs = fix_n_wrapper_jobs

        self.check_args()


class Feat_Selector(Piece):

    def __init__(self, obj, params=0, scope='all',
                 base_model=None, extra_params=None):
        ''' Feat_Selector is a base piece of :class:`Model_Pipeline`, which is designed
        to preform feature selection.

        Parameters
        ----------
        obj : str
            `obj` selects the feature selection strategy to use.
            See :ref:`Feat Selectors`
            for all avaliable options. Notably, if 'rfe' is passed, then a
            base model must also be passed!

            See :ref:`Pipeline Objects` to read more about pipeline objects
            in general.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` set an associated distribution of hyper-parameters to
            potentially search over with this Feat_Selector.
            Preset param distributions are
            listed for each choice of params with the corresponding
            obj at :ref:`Feat Selectors`,
            and you can read more on how params
            work more generally at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified
            feature selector will have access to.
            See :ref:`Scopes` for more information on how scopes can
            be specified.

            ::

                default = 'all'

        base_model : :class:`Model`, :class:`Ensemble` or None, optional
            If 'rfe' is passed to obj, then a base_model is required in
            order to perform recursive feature elimination.
            The base model can be
            any valid argument accepts by
            param `model` in :class:`Model_Pipeline`.

            ::

                default = None

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.base_model = deepcopy(base_model)
        self.base_model_type = None
        self.extra_params = extra_params
        self.check_args()


class Model(Piece):

    _is_model = True

    def __init__(self, obj, params=0, scope='all', param_search=None,
                 target_scaler=None, extra_params=None):
        ''' Model represents a base components of the :class:`Model_Pipeline`,
        specifically a single Model / estimator.
        Model can also be used as a component
        in building other pieces of the model pipeline,
        e.g., :class:`Ensemble`.

        Parameters
        ----------

        obj : str, or custom obj
            `obj` selects the base model object to use from either
            a preset str indicator found at :ref:`Models`,
            or from a custom passed user model (compatible w/ sklearn api).

            See :ref:`Pipeline Objects` to
            read more about pipeline objects in general.

            `obj` should be wither a single str indicator or a
            single custom model object, and not passed a list-like of either.
            If an ensemble of models is requested, then see :class:`Ensemble`.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` optionally set an
            associated distribution of hyper-parameters to
            this model object. Preset param distributions are
            listed for each choice of obj at :ref:`Models`,
            and you can read more on
            how params work more generally at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified model
            should work on. See :ref:`Scopes` for more
            information on how scopes can
            be specified.

            ::

                default = 'all'

        param_search : :class:`Param_Search`, None, optional
            If None, by default, this will be a base model.
            Alternatively, by passing a :class:`Param_Search` instance here,
            it specifies that this model should be wrapped in a
            Nevergrad hyper-parameter search object.

            This can be useful to create Model's which have
            a nested hyper-parameter tuning independent from the
            other pipeline steps.

            ::

                default = None

        target_scaler : Scaler, None, optional

            Still somewhat experimental, can pass
            a Scaler object here and have this model
            perform target scaling + reverse scaling.

            Note: Has not been fully tested in
            complicated nesting cases, e.g., if Model is
            wrapping a nested Model_Pipeline, this param will
            likely break.

            ::

                default = None

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None

        '''

        self.obj = obj
        self.params = params
        self.scope = scope
        self.param_search = param_search
        self.target_scaler = target_scaler
        self.extra_params = extra_params

        self.check_args()

    def view(self, problem_type='regression'):

        from ..pipeline.Models import get_base_model_and_params, AVALIABLE

        proc_str = proc_type_dep_str(self.obj, AVALIABLE, problem_type)
        model, model_params =\
            get_base_model_and_params(proc_str, self.extra_params,
                                      self.params, True)
        model_params = {'__'.join(m.split('__')[1:]): model_params[m]
                        for m in model_params}

        return model, model_params


class Ensemble(Piece):

    _is_model = True

    def __init__(self, obj, models,
                 params=0, scope='all',
                 param_search=None,
                 target_scaler=None,
                 base_model=None,
                 cv_splits=None,
                 is_des=False,
                 single_estimator=False,
                 des_split=.2,
                 n_jobs_type='ensemble',
                 extra_params=None):
        ''' The Ensemble object is valid base
        :class:`Model_Pipeline` piece, designed
        to be passed as input to the `model` parameter
        of :class:`Model_Pipeline`, or
        to its own models parameters.

        This class is used to create a variety ensembled models,
        typically based on
        :class:`Model` pieces.

        Parameters
        ----------
        obj : str
            Each str passed to ensemble refers to a type of ensemble to train,
            based on also the passed input to the `models` parameter,
            and also the
            additional parameters passed when init'ing Ensemble.

            See :ref:`Ensemble Types` to see all
            avaliable options for ensembles.

            Passing custom objects here, while technically possible,
            is not currently full supported.
            That said, there are just certain assumptions that
            the custom object must meet in order to work, specifially,
            they should have simmilar input params to other simmilar existing
            ensembles, e.g., in the case the `single_estimator` is False
            and `needs_split` is also False, then the passed object needs
            to be able to accept an input parameter `estimators`,
            which accepts a list of (str, estimator) tuples.
            Whereas if needs_split is still False,
            but single_estimator is True, then the passed object needs
            to support an init param of `base_estimator`,
            which accepts a single estimator.

        models : :class:`Model`, :class:`Ensemble` or list of
            The `models` parameter is designed to accept any single model-like
            pipeline parameter object, i.e.,
            :class:`Model` or even another :class:`Ensemble`.
            The passed pieces here will be used along with the
            requested ensemble object to
            create the requested ensemble.

            See :class:`Model` for how to create a valid base model(s)
            to pass as input here.

        params : int, str or dict of :ref:`params<Params>`, optional
            `params` sets as associated distribution of hyper-parameters
            for this ensemble object. These parameters will be used only
            in the context of a hyper-parameter search.
            Notably, these `params` refer to the ensemble obj itself,
            params for base `models` should be passed
            accordingly when creating the base models.
            Preset param distributions are listed at :ref:`Ensemble Types`,
            under each of the options for ensemble obj's.

            You can read more about generally about
            hyper-parameter distributions as associated with
            objects at :ref:`Params`.

            ::

                default = 0

        scope : :ref:`valid scope<Scopes>`, optional
            `scope` determines on which subset of features the specified
            ensemble model
            should work on. See :ref:`Scopes` for more
            information on how scopes can
            be specified.

            ::

                default = 'all'

        param_search : :class:`Param_Search`, None, optional
            If None, by default, this will be a base ensemble model.
            Alternatively, by passing a :class:`Param_Search` instance here,
            it specifies that this model should be wrapped in a
            Nevergrad hyper-parameter search object.

            This can be useful to create Model's which have
            a nested hyper-parameter tuning independent from the
            other pipeline steps.

            ::

                default = None

        target_scaler : Scaler, None, optional

            Still somewhat experimental, can pass
            a Scaler object here and have this model
            perform target scaling + reverse scaling.

            scope in the passed scaler is ignored.

            Note: Has not been fully tested in
            complicated nesting cases, e.g., if Model is
            wrapping a nested Model_Pipeline, this param will
            likely break.

            ::

                default = None

        base_model : :class:`Model`, None, optional
            In the case that an ensemble is passed which has
            the parameter `final_estimator` (not base model!),
            for example in the case of stacking,
            then you may pass a Model type
            object here to be used as that final estimator.

            Otherwise, by default this will be left as None,
            and if the requested ensemble has the final_estimator
            parameter, then it will pass None to the object
            (which is typically for setting the default).

            ::

                default = None

        cv_splits : :class:`CV_Splits` or None, optional
            Used for passing custom CV split behavior to
            ensembles which employ splits, e.g., stacking.

            ::

                default = None

        is_des : bool, optional
            `is_des` refers to if the requested ensemble obj requires
            a further training test split in order to train the base ensemble.
            As of right now, If this parameter is True, it means that the
            base ensemble is from the
            `DESlib library <https://deslib.readthedocs.io/en/latest/>`_ .
            Which means
            the base ensemble obj must have a `pool_classifiers`
            init parameter.

            The following `des_split` parameter determines the
            size of the split if
            is_des is True.

            ::

                default = False

        single_estimator : bool, optional
            The parameter `single_estimator` is used to let the
            Ensemble object know if the `models` must be a single estimator.
            This is used for ensemble types
            that requires an init param `base_estimator`.
            In the case that multiple models
            are passed to `models`, but `single_estimator` is True,
            then the models will automatically
            be wrapped in a voting ensemble,
            thus creating one single estimator.

            ::

                default = False

        des_split : float, optional
            If `is_des` is True, then the passed ensemble must be
            fit on a seperate validation set.
            This parameter determines the size
            of the further train/val split on initial
            training set passed to
            the ensemble. Where the size is comptued as
            the a percentage of the total size.

            ::

                default = .2

        n_jobs_type : 'ensemble' or 'models', optional
            Valid options are either 'ensemble' or 'models'.

            This parameter controls how the total n_jobs are distributed, if
            'ensemble', then the n_jobs will be used all in the ensemble object
            and every instance within the sub-models set to n_jobs = 1.
            Alternatively, if passed 'models', then the ensemble
            object will not be multi-processed, i.e.,
            will be set to n_jobs = 1, and the n_jobs will
            be distributed to each base model.

            If you are training a stacking regressor for example
            with n_jobs = 16, and you have 16+ models, then 'ensemble'
            is likely a good choice here. If instead you have only
            3 base models, and one or more of those 3 could benefit from
            a higher n_jobs, then setting n_jobs_type to 'models' might
            give a speed-up.

            ::

                default = 'ensemble'

        extra_params : :ref`extra params dict<Extra Params>`, optional

            See :ref:`Extra Params`

            ::

                default = None
        '''

        self.obj = obj

        # Force passed models if not a list, into a list
        if not isinstance(models, list):
            models = [models]
        self.models = models

        self.params = params
        self.scope = scope
        self.param_search = param_search
        self.target_scaler = target_scaler
        self.base_model = base_model
        self.cv_splits = cv_splits
        self.is_des = is_des
        self.des_split = des_split
        self.single_estimator = single_estimator
        self.n_jobs_type = n_jobs_type
        self.extra_params = extra_params

        self.check_args()

    def check_extra_args(self):

        if isinstance(self.models, list):
            for model in self.models:
                if not hasattr(model, '_is_model'):
                    raise IOError(
                        'All models must be valid Model/Ensemble !')

        else:
            if not hasattr(self.models, '_is_model'):
                raise IOError(
                    'Passed model in models must be a valid Model/Ensemble.')


class Param_Search(Params):

    def __init__(self, search_type='RandomSearch',
                 splits=3, n_repeats=1,
                 cv='default',
                 n_iter=10,
                 scorer='default',
                 weight_scorer=False,
                 mp_context='default',
                 n_jobs='default',
                 dask_ip=None,
                 memmap_X=False,
                 search_only_params=None,
                 CV='depreciated',
                 _random_state=None,
                 _splits_vals=None,
                 _cv=None,
                 _scorer=None,
                 _n_jobs=None):
        ''' Param_Search is special input object designed to be
        used with :class:`Model_Pipeline`.
        Param_Search defines a hyperparameter search strategy.
        When passed to :class:`Model_Pipeline`,
        its search strategy is applied in the context of any set :ref:`Params`
        within the base pieces.
        Specifically, there must be atleast one parameter search
        somewhere in the object Param_Search is passed!

        All backend hyper-parameter searches make use of the
        <https://github.com/facebookresearch/nevergrad>`_ library.

        Parameters
        ----------
        search_type : str, optional
            The type of nevergrad hyper-parameter search to conduct. See
            :ref:`Search Types<Search Types>` for all avaliable options.
            Also you may further look into
            nevergrad's experimental varients if you so choose,
            this parameter can accept
            those as well.

            New: You may pass 'grid' here in addition to the
            supported nevergrad searches. This will use sklearn's
            GridSearch. Note in this case some of the other parameters
            are ignored, these are: weight_scorer, mp_context, dask_ip,
            memmap_X, search_only_params

            ::

                default = 'RandomSearch'

        splits : int, float, str or list of str, optional
            In order to optimize hyper-parameters, some sort of
            internal cross validation must be specified,
            such that combinations of hyper-parameters can be evaluated
            on different data then they were trained on.
            `splits` allows you to specify the base of what CV strategy
            should be used to evaluate every `n_iter` combination
            of hyper-parameters.

            Specifically, options for split are:

            - int
                The number of k-fold splits to conduct. (E.g., 3 for
                3-fold CV split to be conducted at
                every hyper-param evaluation).

            - float
                Must be 0 < `splits` < 1, and defines a
                single train-test like split,
                with `splits` % of the current training data size used
                as a validation set.

            - str
                If a str is passed, then it must correspond to a
                loaded Strat variable. In
                this case, a leave-out-group CV will be used according
                to the value of the
                indicated Strat variable (E.g., a leave-out-site CV scheme).

            - list of str
                If multiple str passed, first determine the overlapping
                unique values from
                their corresponing loaded Strat variables,
                and then use this overlapped
                value to define the leave-out-group CV as described above.

            Also note that `n_repeats` will work with any of these options,
            but say in the case of
            a leave out group CV, would be awfully redundant,
            versus, with a passed float value, very reasonable.

            ::

                default = 3

        n_repeats : int, optional
            Given the base hyper-param search CV defined /
            described in the `splits` param, this
            parameter further controls if the defined train/val splits
            should be repeated (w/ different random
            splits in all cases but the leave-out-group passed str option).

            For example, if `n_repeats` is set to 2, and `splits` is 3,
            then a twice repeated 3-fold CV
            will be performed to evaluate every choice of `n_iter`
            hyper-params.

            ::

                default = 1

        cv : :class:`CV` or 'default', optional
            If left as default 'default', use the class defined CV behavior
            for the splits, otherwise can pass custom behavior.

            ::

                default = 'default'

        n_iter : int, optional
            The number of hyper-parameters to try / budget
            of the underlying search algorithm.
            How well a hyper-parameter search works and how long
            it takes will be very dependent on this parameter
            and the defined internal CV strategy
            (via `splits` and `n_repeats`). In general, if too few
            choices are provided
            the algorithm will likely not select high
            performing hyper-paramers, and alternatively
            if too high a value/budget is
            set, then you may find overfit/non-generalize
            hyper-parameter choices. Other factors which will influence
            the 'right' number of `n_iter` to specify are:

            - `search_type`
                Depending on the underlying search type, it may
                take a bigger or smaller budget
                on average to find a good set of hyper-parameters

            - The dimension of the underlying search space
                If you are only optimizing a few, say 2,
                underlying parameter distributions,
                this will require a far smaller budget then say a really
                high dimensional search space.

            - The CV strategy
                The CV strategy defined via `splits` and `n_repeats`
                may make it easier or harder to overfit when searching
                for hyper-parameters, thus conceptually
                a good choice of CV strategy can serve to increase
                the number `n_iter` you can use
                before overfitting, or conversely a bad choice may limit it.

            - Number of data points / subjects
                Along with CV strategy, the number of data points/subjects
                will greatly influence
                how quickly you overfit, and therefore a
                good choice of `n_iter`.

            Notably, one can always if they have the resources
            simply expiriment with this parameter.

            ::

                default = 10

        scorer : str or 'default', optional
            In order for a set of hyper-parameters to be evaluated,
            a single scorer must be defined.

            For a full list of supported scorers please view the
            scikit-learn docs at:
            https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

            If left as 'default', assign a reasonable scorer based on the
            passed problem type.

            - 'regression'  : 'explained_variance'
            - 'binary'      : 'matthews'
            - 'categorical' : 'matthews'

            Be careful to make sure to select an appropriate scorer for
            the problem type.

            Only one value of `scorer` may be passed here.

            ::

                default = 'default'

        weight_scorer : bool or 'default', optional
            `weight_scorer` describes if the scorer of interest
            should be weighted by
            the number of subjects within each validation fold.
            So, for example, if
            a leave-out-group CV scheme is specified to `splits`,
            and the groups have
            drastically different numbers of subjects,
            then you may want to consider
            weighting the final average validation metric
            (as computed across in this case
            all groups used by themselves) by the number
            of subjects in each fold.

            ::

                default = False

        mp_context : str, optional
            When a hyper-parameter search is launched, there are different
            ways through python that the multi-processing can be launched
            (assuming n_jobs > 1). Occassionally some choices can lead to
            unexpected errors.

            Choices are:
            - 'default': If 'default' use the BPt mp_context.

            - 'loky': Create and use the python library loky backend.

            - 'fork': Python default fork mp_context

            - 'forkserver': Python default forkserver mp_context

            - 'spawn': Python default spawn mp_context

            ::

                default = 'default'

        n_jobs : int or 'default', optional
            The number of cores to be used for the
            search. In general, this parameter
            should be left as 'default', which will set it
            based on the n_jobs as set in the problem spec-
            and will attempt to automatically change this
            value if say in the context of nesting.

            ::

                default = 'default'

        dask_ip : str or None, optional
            If None, default, then ignore this parameter..

            For experimental Dask support.
            This should be the ip of a created dask
            cluster. A dask Client object will be created
            and passed this ip in order to connect to the cluster.

            ::

                default = None

        memmap_X : bool, optional
            When passing large memory arrays in each parameter search,
            it can be useful as a memory reduction technique to pass
            numpy memmap'ed arrays. This solves an issue where
            the loky backend will not properly pass too large arrays.

            Warning: This can slow down code, and only reduces the actual
            memory consumption of each job by a little bit.

            Note: If passing a dask_ip, this option will be skipped,
            as if using the dask backend, then large X's will be
            pre-scattered instead.

            ::

                default = False

        search_only_params : dict or None, optional
            In some rare cases, it may be the case that you want
            to specify that certain parameters be passed only during
            the nested parameter searches. A dict of parameters
            can be passed here to accomplish that. For example,
            if passing:

            search_only_params = {'svm classifier__probability': False}

            And assuming that the default / selecting parameter for this svm
            classifier for probaility is True by default,
            then only when exploring
            nested hyper-parameter options will
            probability be set to False, but
            when fitting the final model with the best
            parameters found from the search,
            it will revert back to the default,
            i.e., in this case probability = True.

            Note: this may be a little bit tricky to use as
            you need to know how to represent the parameters correctly!

            To ignore this parameter / option.
            simply keep the default value of None

            ::

                default = None

        CV : 'depreciated'
            Switching to passing cv parameter as cv instead of CV.
            Will raise error if anything is passed here.

            ::

                default = 'depreciated'
        '''

        self.search_type = search_type

        self.splits = splits
        self.n_repeats = n_repeats
        self.cv = cv
        self.n_iter = n_iter
        self.scorer = scorer
        self.weight_scorer = weight_scorer
        self.mp_context = mp_context
        self.n_jobs = n_jobs
        self.dask_ip = dask_ip
        self.memmap_X = memmap_X

        if search_only_params is None:
            search_only_params = {}
        self.search_only_params = search_only_params

        self._random_state = _random_state
        self._splits_vals = _splits_vals
        self._cv = _cv
        self._scorer = _scorer
        self._n_jobs = _n_jobs

        if CV != 'depreciated':
            raise RuntimeError('Pass as cv instead of CV!')

        self.CV = CV

        self.check_args()

    def set_random_state(self, random_state):
        self._random_state = random_state

    def set_n_jobs(self, n_jobs):

        if self.n_jobs == 'default':
            self._n_jobs = n_jobs
        else:
            self._n_jobs = self.n_jobs

    def set_scorer(self, problem_type):
        self._scorer =\
            process_scorers(self.scorer,
                            problem_type)[2]

    def set_cv(self, cv):
        self._cv = cv

    def set_split_vals(self, vals):
        self._splits_vals = vals

    def check_args(self):

        if isinstance(self.scorer, list):
            raise IOError('scorer within Param Search cannot be list-like')
        if isinstance(self.weight_scorer, list):
            raise IOError(
                'weight_scorer within Param Search cannot be list-like')


class Shap_Params(Params):

    def __init__(self,
                 avg_abs=False,
                 linear_feature_perturbation="interventional",
                 linear_nsamples=1000,
                 tree_feature_perturbation='tree_path_dependent',
                 tree_model_output='raw',
                 tree_tree_limit=None,
                 kernel_nkmean=20,
                 kernel_link='default',
                 kernel_nsamples='auto',
                 kernel_l1_reg='auto'):
        '''
        There are a number of parameters associated
        with using shap to determine
        feature importance. The best way to understand Shap
        is almost certainly through
        their documentation directly, `Shap Docs <https://shap.readthedocs.io/en/latest/>`_
        Just note when using Shap within BPt to pay attention
        to the version on the shap
        documentation vs. what is currently supported within BPt.

        Broadly, shap feature importance params are
        split up into if they are used
        to explain linear models, tree based models
        or an abitrary model (kernel).
        Be warned, kernel shap generally takes a long time to run.

        The type of shap to use will be automatically
        computed based on the Model to
        explain, but the parameter below are still
        split up by type, i.e., with the
        type prended before the parameter name.

        (A number of the params below are copy-pasting from the
        relevant Shap description)

        Parameters
        ----------
        avg_abs : bool, optional
            This parameter is considered regardless of
            the underlying shap model. If
            set to True, then when computing global feature
            importance from the
            initially computed local shap feature importance
            the average of the absolute value
            will be taken! When set to False, the average value will be taken.
            One might want to
            set this to True if only concerned with the magnitude
            of feature importance, rather
            than the sign.

            ::

                default = False

        linear_feature_perturbation : {"interventional", "correlation_dependent"}, optional
            Only used with linear base models.
            There are two ways we might want to compute SHAP values,
            either the full conditional SHAP
            values or the interventional SHAP values.
            For interventional SHAP values we break any
            dependence structure between features in the model and
            so uncover how the model would behave if we
            intervened and changed some of the inputs. For the full
            conditional SHAP values we respect
            the correlations among the input features, so if the model depends on one input but that
            input is correlated with another input, then both get some credit for the model's behavior. The
            interventional option stays "true to the model" meaning it will only give credit to features that are
            actually used by the model, while the correlation option stays "true to the data" in the sense that
            it only considers how the model would behave when respecting the correlations in the input data.
            For sparse case only interventional option is supported.

            ::

                default = "interventional"

        linear_nsamples : int, optional
            Only used with linear base models.
            Number of samples to use when estimating the transformation matrix used to account for
            feature correlations.

            ::

                default = 1000

        tree_feature_perturbation : {"interventional", "tree_path_dependent"}, optional
            Only used with tree based models.
            Since SHAP values rely on conditional expectations
            we need to decide how to handle correlated
            (or otherwise dependent) input features.
            The "interventional" approach breaks the dependencies between
            features according to the rules dictated by casual
            inference (Janzing et al. 2019). Note that the
            "interventional" option requires a background dataset
            and its runtime scales linearly with the size
            of the background dataset you use. Anywhere from 100 to 1000
            random background samples are good
            sizes to use. The "tree_path_dependent" approach is to just follow
            the trees and use the number
            of training examples that went down each leaf to represent
            the background distribution. This approach
            does not require a background dataset and so is used
            by default when no background dataset is provided.

            ::

                default = "tree_path_dependent"

        tree_model_output : {"raw", "probability", "log_loss"} optional
            Only used with tree based models.
            What output of the model should be explained. If "raw" then we explain the raw output of the
            trees, which varies by model. For regression models "raw" is the standard output, for binary
            classification in XGBoost this is the log odds ratio. If model_output is the name of a supported
            prediction method on the model object then we explain the output of that model method name.
            For example model_output="predict_proba" explains the result of calling model.predict_proba.
            If "probability" then we explain the output of the model transformed into probability space
            (note that this means the SHAP values now sum to the probability output of the model). If "logloss"
            then we explain the log base e of the model loss function, so that the SHAP values sum up to the
            log loss of the model for each sample. This is helpful for breaking down model performance by feature.
            Currently the probability and logloss options are only supported when feature_dependence="independent".

            ::

                default = 'raw'

        tree_tree_limit : None or int, optional
            Only used with tree based models.
            Limit the number of trees used by the model. By default None means no use the limit of the
            original model, and -1 means no limit.

            ::

                default = None

        kernel_n_kmeans : int or None, optional
            Used when the underlying model is not linear or tree based.
            This setting offers a speed up to the kernel estimator by replacing
            the background dataset with a kmeans representation of the data.
            Set this option to None in order to use the full dataset directly,
            otherwise the int passed will the determine 'k' in the kmeans algorithm.

            ::

                default = 20

        kernel_link : {"identity", "logit", "default"}
            Used when the underlying model is not linear or tree based.

            A generalized linear model link to connect the feature importance values to the model
            output. Since the feature importance values, phi, sum up to the model output, it often makes
            sense to connect them to the ouput with a link function where link(outout) = sum(phi).
            If the model output is a probability then the LogitLink link function makes the feature
            importance values have log-odds units.

            If 'default', set to using 'logic' for binary + categorical problems, along
            with explaining predict_proba from the model, and 'identity' and explaining
            predict for regression.

            ::

                default = 'default'

        kernel_nsamples : "auto" or int, optional
            Used when the underlying model is not linear or tree based.
            Number of times to re-evaluate the model when explaining each prediction. More samples
            lead to lower variance estimates of the SHAP values. The "auto" setting uses
            `nsamples = 2 * X.shape[1] + 2048`.

            ::

                default = 'auto'

        kernel_l1_reg : "num_features(int)", "auto", "aic", "bic", or float, optional
            Used when the underlying model is not linear or tree based.
            The l1 regularization to use for feature selection
            (the estimation procedure is based on
            a debiased lasso). The auto option currently uses "aic" when less that 20% of the possible sample
            space is enumerated, otherwise it uses no regularization. THE BEHAVIOR OF "auto" WILL CHANGE
            in a future version to be based on num_features instead of AIC.
            The "aic" and "bic" options use the AIC and BIC rules for regularization.
            Using "num_features(int)" selects a fix number of top features. Passing a float directly sets the
            "alpha" parameter of the sklearn.linear_model.Lasso model used for feature selection.

            ::

                default = 'auto'

        '''

        try:
            import shap
        except ImportError:
            raise ImportError('You must have shap installed to use shap')

        self.avg_abs = avg_abs
        self.linear_feature_perturbation = linear_feature_perturbation
        self.linear_nsamples = linear_nsamples
        self.tree_feature_perturbation = tree_feature_perturbation
        self.tree_model_output = tree_model_output
        self.tree_tree_limit = tree_tree_limit
        self.kernel_nkmean = kernel_nkmean
        self.kernel_link = kernel_link
        self.kernel_nsamples = kernel_nsamples
        self.kernel_l1_reg = kernel_l1_reg


class Feat_Importance(Params):

    def __init__(self, obj, scorer='default',
                 shap_params='default', n_perm=10,
                 inverse_global=False, inverse_local=False):
        '''
        There are a number of options for creating Feature Importances in BPt.
        See :ref:`Feat Importances` to learn more about
        feature importances generally.
        The way this object works, is that you can a type
        of feature importance, and then
        its relevant parameters. This object is designed
        to passed directly to
        :class:`Model_Pipeline`.

        Parameters
        ----------
        obj : str
            `obj` is the str indiciator for which feature importance to use.
            See :ref:`Feat Importances` for what options are avaliable.

        scorer : str or 'default', optional

            If a permutation based feature importance is being used, then a scorer is
            required.

            For a full list of supported scorers please view the scikit-learn docs at:
            https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

            If left as 'default', assign a reasonable scorer based on the
            passed problem type.

            - 'regression'  : 'explained_variance'
            - 'binary'      : 'matthews'
            - 'categorical' : 'matthews'

            ::

                default = 'default'

        shap_params : :class:`Shap_Params` or 'default', optional
            If a shap based feature importance is used, it is neccicary to define
            a number of relevant parameters for how the importances should be calculated.
            See :class:`Shap_Params` for what these parameters are.

            If 'default' is passed, then shap_params will be set to either the default values of
            :class:`Shap_Params` if shap feature importances are being used, or None if not.

            ::

                default = 'default'

        n_perm : int, optional
            If a permutation based feature importance method is selected, then
            it is neccicary to indicate how many random permutations each feature 
            should be permuted.

            ::

                default = 10

        inverse_global : bool
            Warning: This feature, along with inverse_local, is still
            expirimental.

            If there are any loaders, or transformers specified
            in the Model_Pipeline,
            then feature importance becomes slightly trickier.
            For example, if you have
            a PCA transformer, and what to calculate averaged feature importance
            across 3-folds, there is no gaurentee 'pca feature 1' is the same from one
            fold to the next. In this case, if set to True, global feature
            importances will be inverse_transformed back into their original feature
            space - per fold. Note: this will only work if all transformers / loaders
            have an implemented reverse_transform function, if one does not for transformer, then
            it will just return 0 for that feature. For a loader w/o, then it will return 
            'No inverse_transform'.

            There are also other cases where this might be a bad idea, for example
            if you are using one hot encoders in your transformers then trying to 
            reverse_transform
            feature importances will yield nonsense (NaN's).

            ::

                default = False

        inverse_local : bool
            Same as inverse_global, but for local feature importances.
            By default this is set to False, as it is
            more memory and computationally expensive to
            inverse_transform this case.

            ::

                default = False

        '''

        self.obj = obj
        self.scorer = scorer

        if shap_params == 'default':

            if 'shap' in self.obj:
                shap_params = Shap_Params()
            else:
                shap_params = None

        self.shap_params = shap_params
        self.n_perm = n_perm
        self.inverse_global = inverse_global
        self.inverse_local = inverse_local

        # For compatibility
        self.params = 0


class Model_Pipeline(Params):

    def __init__(self,
                 loaders=None, imputers='default',
                 scalers=None, transformers=None,
                 feat_selectors=None,
                 model='default',
                 param_search=None,
                 n_jobs='default',
                 cache='depreciated',
                 feat_importances='depreciated'):
        ''' Model_Pipeline is defined as essentially a wrapper around
        all of the explicit modelling pipeline parameters. This object is
        used as input to
        :func:`Evaluate <BPt.BPt_ML.Evaluate>`
        and :func:`Test <BPt.BPt_ML.Test>`

        The ordering of the parameters listed below defines the pre-set
        order in which these Pipeline pieces are composed
        (params up to model, param_search is not an ordered pipeline piece).
        For more flexibility, one can always use custom defined objects,
        or even pass custom defined
        pipelines directly to model
        (i.e., in the case where you have a specific pipeline you want to use
        already defined, but say just want to use the loaders from BPt).

        Parameters
        ----------
        loaders : :class:`Loader`, list of or None, optional
            Each :class:`Loader` refers to transformations
            which operate on loaded Data_Files
            (See :func:`Load_Data_Files <BPt.BPt_ML.Load_Data_Files>`).
            See :class:`Loader`
            explcitly for more information on how to create a valid object,
            with relevant params and scope.

            In the case that a list of Loaders is passed to loaders,
            if a native python list, then passed loaders will
            be applied sequentially (likely each
            passed loader given a seperate scope, as the output from
            one loader cannot be input
            to another- note to create actual sequential loader steps,
            look into using the
            :class:`Pipe` wrapper
            argument when creating a single :class:`Loader` obj).

            Passed loaders can also be wrapped in a
            :class:`Select` wrapper, e.g., as either

            .. code-block::

                # Just passing select
                loaders = Select([Loader(...), Loader(...)])

                # Or nested
                loaders = [Loader(...), Select([Loader(...), Loader(...)])]

            In this way, most of the pipeline objects can accept lists,
            or nested lists with
            param wrapped, not just loaders!

            .. code-block::

                default = None

        imputers : :class:`Imputer`, list of or None, optional
            If there is any missing data (NaN's) that have been kept
            within data or covars, then an imputation strategy must be
            defined! This param controls what kind of
            imputation strategy to use.

            Each :class:`Imputer` contains information
            around which imputation
            strategy to use, what scope it is applied
            to (in this case only 'float' vs. 'cat'),
            and other relevant base parameters
            (i.e., a base model if an iterative imputer is selected).

            In the case that a list of :class:`Imputer` are passed,
            they will be applied sequentially,
            though note that unless custom scopes
            are employed, at most passing only an imputer for float data and
            an imputer for categorical data makes sense.
            You may also use input wrapper, like :class:`Select`.

            In the case that no NaN data is passed, but imputers is not None,
            it will simply be set to None.

            ::

                default = [Imputer('mean', scope='float'),
                           Imputer('median', scope='cat')]

        scalers : :class:`Scaler`, list of or None, optional
            Each :class:`Scaler` refers to any potential data scaling where a
            transformation on the data
            (without access to the target variable) is
            computed, and the number of features or
            data points does not change.
            Each :class:`Scaler` object contains information
            about the base object, what
            scope it should be applied to, and saved param
            distributions if relevant.

            As with other pipeline params, scalers can
            accept a list of :class:`Scaler` objects,
            in order to apply sequential transformations
            (or again in the case where each object has a seperate scope,
            these are essentially two different streams of transformations,
            vs. when two Scalers with the same scope are passed,
            the output from one
            is passed as input to the next). Likewise,
            you may also use valid input wrappers,
            e.g., :class:`Select`.

            By default no scaler is used, though it is reccomended.

            ::

                default = None

        transformers : :class:`Transformer`, list of or None, optional
            Each :class:`Transformer` defines a type of transformation to
            the data that changes the number of
            features in perhaps non-deterministic
            or not simply removal
            (i.e., different from feat_selectors), for example
            applying a PCA, where both the number of features change, but
            also the new features do not 1:1 correspond to the original
            features. See :class:`Transformer` for more information.

            Transformers can be composed sequentially with list or special
            input type wrappers, the same as other objects.

            ::

                default = None

        feat_selectors : :class:`Feat_Selector`, list of or None, optional
            Each :class:`Feat_Selector` refers to an optional
            feature selection stage
            of the Pipeline. See :class:`Feat_Selector` for specific options.

            Input can be composed in a list, to apply feature selection sequentially,
            or with special Input Type wrapper, e.g., :class:`Select`.

            ::

                default = None

        model : :class:`Model`, :class:`Ensemble`, optional
            model accepts one input of type
            :class:`Model` or :class:`Ensemble`. Though,
            while it cannot accept a list (i.e., no sequential behavior), you
            may still pass Input Type wrapper like :class:`Select` to perform
            model selection via param search.

            See :class:`Model` for more information on how to specify a single
            model to BPt, and :class:`Ensemble` for information on how
            to build an ensemble of models.

            Note: You must have provide a model, there is
            no option for None. Instead
            default behavior is to use a ridge regression.

            ::

                default = Model('ridge')

        param_search : :class:`Param_Search` or None, optional
            :class:`Param_Search` can be provided in order
            to specify a corresponding
            hyperparameter search for the provided pipeline pieces.
            When defining each
            piece, you may set hyperparameter distributions for that piece.
            If param search
            is None, these distribution will be essentially ignored,
            but if :class:`Param_Search`
            is passed here, then they will be used along with the
            strategy defined in the passed
            :class:`Param_Search` to conduct a nested hyper-param search.

            Note: If using input wrapper types like :class:`Select`,
            then a param search must be passed!

            ::

                default = None

        n_jobs : int or 'default', optional
            The number of cores to be used with this pipeline.
            In general, this parameter
            should be left as 'default', which will set it
            based on the n_jobs as set in the problem spec-
            and will attempt to automatically change this
            value if say in the context of nesting.

            ::

                default = 'default'

        cache : depreciated
            The cache parameter has been depreciated,
            use the cache_loc params within individual pieces instead.

            ::

                default = 'depreciated'


        feat_importances : depreciated
            Feature importances in a past version of BPt were
            specified via this Model Pipeline object.
            Now they should be provided to either
            :func:`Evaluate <BPt.BPt_ML.Evaluate>`
            and :func:`Test <BPt.BPt_ML.Test>`

        ::

            default = 'depreciated'

        '''

        if isinstance(loaders, str):
            loaders = Loader(loaders)
        self.loaders = loaders

        if imputers == 'default':
            imputers = [Imputer('mean', scope='float'),
                        Imputer('median', scope='cat')]
            print('Passed default imputers, setting to:', imputers)
        elif isinstance(imputers, str):
            imputers = Imputer(imputers)
        self.imputers = imputers

        if isinstance(scalers, str):
            scalers = Scaler(scalers)
        self.scalers = scalers

        if isinstance(transformers, str):
            transformers = Transformer(transformers)
        self.transformers = transformers

        if isinstance(feat_selectors, str):
            feat_selectors = Feat_Selector(feat_selectors)
        self.feat_selectors = feat_selectors

        if model == 'default':
            model = Model('ridge')
            print('Passed default model, setting to:', print(model))
        elif isinstance(model, str):
            model = Model(model)
        self.model = model

        self.param_search = param_search
        self.n_jobs = n_jobs

        if cache != 'depreciated':
            print('Warning: Passing cache is depreciated.')
        self.cache = cache

        if feat_importances != 'depreciated':
            print('Warning: Passing feature importances have been moved ',
                  'to the Evaluate and Test functions!')

        # Regardless save the value to avoid sklearn warnings
        self.feat_importances = feat_importances

        # Perform all preproc on input which can be run
        # more then once, these are essentially checks on the input
        self._proc_checks()

    def _proc_all_pieces(self, func):
        '''Helper to proc all pieces with a function
        that accepts a list of pieces'''

        for param_name in ORDERED_NAMES:
            params = getattr(self, param_name)

            if params is None:
                new_params = params

            elif isinstance(params, list):
                new_params = func(params)

            # If not list-
            else:
                new_params = func([params])
                if isinstance(new_params, list) and len(new_params) == 1:
                    new_params = new_params[0]

            setattr(self, param_name, new_params)

    def _proc_duplicates(self, params):

        if isinstance(params, list):
            new_params = deepcopy(params)
            del new_params[:]

            if len(params) == 1:
                return self._proc_duplicates(params[0])

            for p in params:
                new_params.append(self._proc_duplicates(p))

            return new_params

        elif hasattr(params, 'scope'):
            scopes = getattr(params, 'scope')
            new_params = []

            if is_duplicate(scopes):
                for scope in scopes:
                    new = deepcopy(params)
                    new.scope = scope
                    new_params.append(new)

                return new_params

        return params

    def _proc_input(self, params):

        for p in params:
            if isinstance(p, list):
                self._proc_input(p)
            else:
                proc_all(p)

        return params

    def _check_args(self, params):

        for p in params:
            if isinstance(p, list):
                self._check_args(p)
            else:
                p.check_args()

        return params

    def check_imputer(self, data):

        # If no NaN, no imputer
        if not pd.isnull(data).any().any():
            self.imputers = None

    def check_samplers(self, func):
        if self.samplers is not None:
            if isinstance(self.samplers, list):
                for s in self.samplers:
                    s.add_strat_u_name(func)
            else:
                self.samplers.add_strat_u_name(func)

    def preproc(self, n_jobs):

        # Base proc checks
        self._proc_checks()

        # Store n_jobs
        self.set_n_jobs(n_jobs)

    def _proc_checks(self):

        try:
            self.model = self.models
            print('Passed models, set as model')
        except AttributeError:
            pass

        try:
            self.transformers = self.transformer
            print('Passed transformer, set as transformers')
        except AttributeError:
            pass

        try:
            self.imputers = self.imputer
            print('Passed imputer, set as imputers')
        except AttributeError:
            pass

        try:
            self.feat_selectors = self.feat_selector
            print('Passed feat_selector, set as feat_selectors')
        except AttributeError:
            pass

        try:
            self.scalers = self.scaler
            print('Passed scaler, set as scalers')
        except AttributeError:
            pass

        try:
            self.loaders = self.loader
            print('Passed loader, set as loaders')
        except AttributeError:
            pass

        to_check = ['ensemble', 'scorer', 'metric',
                    'feature_selector', 'feature_selectors']
        for p in to_check:
            if hasattr(self, p):
                print('Warning: Model_Pipeline user set param', p,
                      ' was set,',
                      'but will have no effect as it is not',
                      ' a valid parameter!')

        # Check for duplicate scopes
        self._proc_all_pieces(self._proc_duplicates)

        # Proc input
        self._proc_all_pieces(self._proc_input)
        proc_all(self.param_search)

        # Double check input args in case something changed
        self._proc_all_pieces(self._check_args)

        # Proc param search if not None
        if self.param_search is not None:
            self.param_search.check_args()

    def set_n_jobs(self, n_jobs):

        if self.n_jobs == 'default':

            # If no param search, each objs base n_jobs is
            # the passed n_jobs
            if self.param_search is None:
                self.n_jobs = n_jobs

            # Otherwise, base jobs are 1
            else:
                self.n_jobs = 1

    def get_ordered_pipeline_params(self):

        # Conv all to list & return in order as a deep copy
        return deepcopy([conv_to_list(getattr(self, piece_name))
                         for piece_name in ORDERED_NAMES])

    def get_indent(self, indent):

        if indent is None:
            return ''
        else:
            ind = 0
            for s in self._p_stack:
                if s == ']':
                    ind += 1
                elif s == '])':
                    ind += 8

            return ' ' * ind

    def params_print(self, params, indent, _print=print, end='\n'):

        if not isinstance(params, list):
            _print(self.get_indent(indent), params, sep='', end=end)
            return

        elif len(params) == 1:
            self.params_print(params[0], 0, _print=_print)
            return

        elif not is_special(params):

            _print(self.get_indent(indent), '[', sep='', end='')
            self._p_stack.append(']')
            self.params_print(params[0], None, _print=_print, end='')
            _print(',', sep='')

        elif is_select(params):
            _print(self.get_indent(indent), 'Select([', sep='', end='')
            self._p_stack.append('])')
            self.params_print(params[0], None, _print=_print, end='')
            _print(',', sep='')

        for param in params[1:-1]:
            self.params_print(param, 0, _print=_print, end='')
            _print(',', sep='')

        self.params_print(params[-1], 0, _print=_print, end='')
        _print(self._p_stack.pop(), end=end)

        return

    def print_all(self, _print=print):

        self._p_stack = []

        _print('Model_Pipeline')
        _print('--------------')

        pipeline_params = self.get_ordered_pipeline_params()
        for name, params in zip(ORDERED_NAMES, pipeline_params):

            if params is not None:
                _print(name + '=\\')
                self.params_print(params, 0, _print=_print)
                _print()

        _print('param_search=\\')
        _print(self.param_search)
        _print()

class Problem_Spec(Params):

    def __init__(self, problem_type='default',
                 target=0, scorer='default', weight_scorer=False,
                 scope='all', subjects='all',
                 n_jobs='default',
                 random_state='default'):
        '''Problem Spec is defined as an object of params encapsulating the set of
        parameters shared by modelling class functions
        :func:`Evaluate <BPt.BPt_ML.Evaluate>` and :func:`Test <BPt.BPt_ML.Test>`

        Parameters
        ----------
        problem_type : str or 'default', optional
            This parameter controls what type of machine learning
            should be conducted. As either a regression, or classification
            where 'categorical' represents a special case of binary classification,
            where typically a binary classifier is trained on each class.

            - 'default'
                Determine the problem type based on how
                the requested target variable is loaded.

            - 'regression', 'f' or 'float'
                For ML on float/continuous target data.

            - 'binary' or 'b'
                For ML on binary target data.

            - 'categorical' or 'c'
                For ML on categorical target data, as multiclass.

            ::

                default = 'default'

        target : int or str, optional
            The loaded target in which to use during modelling.
            This can be the int index (as assigned by order loaded in,
            e.g., first target loaded is 0, then the next is 1),
            or the name of the target column.
            If only one target is loaded, just leave as default of 0.

            ::

                default = 0

        scorer : str or list, optional
            Indicator str for which scorer(s) to use when calculating
            average validation score in Evaluate, or Test set score in Test.

            A list of str's can be passed as well, in this case, scores for
            all of the requested scorers will be calculated and returned.

            Note: If using a Param_Search, the Param_Search object has a
            scorer parameter as well. This scorer describes
            the scorer optimized
            in a parameter search.

            For a full list of supported scorers please view the
            scikit-learn docs at:
            https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

            If left as 'default', assign a reasonable scorer based on the
            passed problem type.

            - 'regression'  : ['explained_variance', 'neg_mean_squared_error']
            - 'binary'      : ['matthews', 'roc_auc', 'balanced_accuracy']
            - 'categorical' : ['matthews', 'roc_auc_ovr', 'balanced_accuracy']

            ::

                default = 'default'

        weight_scorer : bool, list of, optional
            If True, then the scorer of interest will be weighted within
            each repeated fold by the number of subjects in that
            validation set.
            This parameter only typically makes sense for
            custom split behavior where
            validation folds may end up with differing sizes.
            When default CV schemes are employed,
            there is likely no point in
            applying this weighting, as the validation
            folds will have simmilar sizes.

            If you are passing mutiple scorers, then you can also pass a 
            list of values for weight_scorer, with each value
            set as boolean True or False,
            specifying if the corresponding scorer by index
            should be weighted or not.

            ::

                default = False

        scope : key str or Scope obj, optional
            This parameter allows the user to optionally
            run an expiriment with just a subset of the loaded features
            / columns.

            See :ref:`Scopes` for a more detailed explained / guide on how scopes
            are defined and used within BPt.

            ::

                default = 'all'

        subjects : str, array-like or Value_Subset, optional
            This parameter allows the user to optionally run Evaluate or Test
            with just a subset of the loaded subjects. It is notably distinct
            from the `train_subjects`, and `test_subjects` parameters directly
            avaliable to Evaluate and Test, as those parameters typically refer
            to train/test splits. Specifically, any value specified for this 
            subjects parameter will be applied AFTER selecting the relevant train
            or test subset.

            One use case for this parameter might be specifying subjects of just one
            sex, where you would still want the same training set for example, but just want
            to test sex specific models.

            If set to 'all' (as is by default), all avaliable subjects will be
            used.

            `subjects` can accept either a specific array of subjects,
            or even a loc of a text file (formatted one subject per line) in
            which to read from.

            A special wrapper, Value_Subset,
            can also be used to specify more specific,
            specifically value specific, subsets of subjects to use.
            See :class:`Value_Subset` for how this input wrapper can be used.

            ::

                default = 'all'

        n_jobs : int, or 'default'
            n_jobs are employed witin the context
            of a call to Evaluate or Test.
            If left as default, the class wide BPt value will be used.

            In general, the way n_jobs are propegated to the different pipeline
            pieces on the backend is that,
            if there is a parameter search, the base
            ML pipeline will all be set to use 1 job,
            and the n_jobs budget will be used
            to train pipelines in parellel to explore different params.
            Otherwise, if no param search,
            n_jobs will be used for each piece individually,
            though some might not support it.

            ::

                default = 'default'

        random_state : int, RandomState instance, None or 'default', optional
            Random state, either as int for a specific seed, or if None then
            the random seed is set by np.random.

            This parameter is used to ensure replicability
            of expirements (wherever possible!).
            In some cases even with a random seed, depending on
            the pipeline pieces being used,
            if any have a component that occassionally yields
            different results, even with the same
            random seed, e.g., some model optimizations,
            then you might still not get
            exact replicicability.

            If 'default', use the saved class value.
            (Defined in :class:`ML <BPt.BPt_ML>`)

            ::

                default = 'default'

        '''

        self.problem_type = problem_type
        self.target = target
        self.scorer = scorer
        self.weight_scorer = weight_scorer
        self.scope = scope
        self.subjects = subjects
        self.n_jobs = n_jobs
        self.random_state = random_state

        self._final_subjects = None

        self._proc_checks()

    def _proc_checks(self):
        proc_all(self)

    def set_final_subjects(self, final_subjects):
        self._final_subjects = final_subjects

    def get_model_spec(self):

        return {'problem_type': self.problem_type,
                'random_state': self.random_state}

    def print_all(self, _print=print):

        _print('Problem_Spec')
        _print('------------')
        _print('problem_type =', self.problem_type)
        _print('target =', self.target)
        _print('scorer =', self.scorer)
        _print('weight_scorer =', self.weight_scorer)
        _print('scope =', self.scope)

        if is_value_subset(self.subjects) or len(self.subjects) < 50:
            _print('subjects =', self.subjects)

        if self._final_subjects is not None:
            _print('len(subjects) =', len(self._final_subjects),
                   '(before overlap w/ train/test subjects)')
        _print('n_jobs =', self.n_jobs)
        _print('random_state =', self.random_state)
        _print()


class CV(Params):

    def __init__(self, groups=None, stratify=None,
                 train_only_loc=None, train_only_subjects=None):
        ''' This objects is used to encapsulate a set of parameters
        for a CV strategy.

        Parameters
        ----------
        groups : str, list or None, optional
            In the case of str input, will assume the str to refer
            to a column key within the loaded strat data,
            and will assign it as a value to preserve groups by
            during any train/test or K-fold splits.
            If a list is passed, then each element should be a str,
            and they will be combined into all unique
            combinations of the elements of the list.

        ::

            default = None

        stratify : str, list or None, optional
            In the case of str input, will assume the str to refer
            to a column key within the loaded strat data, or a loaded target col.,
            and will assign it as a value to preserve
            distribution of groups by during any train/test or K-fold splits.
            If a list is passed, then each element should be a str,
            and they will be combined into all unique combinations of
            the elements of the list.

            Any target_cols passed must be categorical
            or binary, and cannot be
            float. Though you can consider loading in a
            float target as a strat,
            which will apply a specific k_bins, and then be valid here.

            In the case that you have a loaded strat val with the same name
            as your target, you can distinguish between the two by passing
            either the raw name, e.g., if they are both loaded as 'Sex',
            passing just 'Sex', will try to use the loaded target. If instead
            you want to use your loaded strat val
            with the same name - you have
            to pass 'Sex' + self.self.strat_u_name
            (by default this is '_Strat').

            ::

                default = None

        train_only_loc : str, Path or None, optional
            Location of a file to load in train_only subjects,
            where any subject loaded as train_only will be assigned to
            every training fold, and never to a testing fold.
            This file should be formatted as one subject per line.

            You can load from a loc and pass subjects, the subjects
            from each source will be merged.

            This parameter is compatible with groups / stratify.

            ::

                default = None

        train_only_subjects : set, array-like, 'nan', or None, optional
            An explicit list or array-like of train_only subjects, where
            any subject loaded as train_only will be assigned to every training
            fold, and never to a testing fold.

            You can also optionally specify 'nan' as input, which
            will add all subjects with any NaN data to train only.

            If you want to add both all the NaN subjects and custom
            subjects, call :func:`Get_Nan_Subjects` to get all NaN subjects,
            and then merge them yourself with any you want to pass.

            You can load from a loc and pass subjects, the subjects
            from each source will be merged.

            This parameter is compatible with groups / stratify.

            ::

                default = None
        '''

        self.groups = groups
        self.stratify = stratify
        self.train_only_loc = train_only_loc
        self.train_only_subjects = train_only_subjects


class CV_Split(Params):

    def __init__(self, cv='default', split=.2,
                 _cv=None, _random_state=None):

        self.cv = cv
        self.split = split
        self._cv = _cv
        self._random_state = _random_state

    def setup(self, cv, random_state):

        self._cv = cv
        self._random_state = random_state

    def get_split(self, train_data_index):

        return self._cv.train_test_split(subjects=train_data_index,
                                         test_size=self.split,
                                         random_state=self._random_state,
                                         return_index=True)

class CV_Splits(Params):

    def __init__(self, cv='default', splits=3, n_repeats=1,
                 _cv=None, _random_state=None, _splits_vals=None):
        ''' This object is used to wrap around a CV strategy at a higher level.

        Parameters
        ----------
        cv: 'default' or :class:`CV`, optional
            If left as default 'default', use the class defined CV behavior
            for the splits, otherwise can pass custom behavior.

        splits : int, float, str or list of str, optional
            `splits` allows you to specify the base of what CV strategy
            should be used.

            Specifically, options for split are:

            - int
                The number of k-fold splits to conduct. (E.g., 3 for
                3-fold CV split to be conducted at
                every hyper-param evaluation).

            - float
                Must be 0 < `splits` < 1, and defines a
                single train-test like split,
                with `splits` % of the current training data size used
                as a validation set.

            - str
                If a str is passed, then it must correspond to a
                loaded Strat variable. In
                this case, a leave-out-group CV will be used according
                to the value of the
                indicated Strat variable (E.g., a leave-out-site CV scheme).

            - list of str
                If multiple str passed, first determine the overlapping
                unique values from
                their corresponing loaded Strat variables,
                and then use this overlapped
                value to define the leave-out-group CV as described above.

            `n_repeats` is designed to work with any of these choices.

            ::

                default = 3

        n_repeats : int, optional
            The number of times to repeat the defined strategy
            as defined in `splits`.

            For example, if `n_repeats` is set to 2, and `splits` is 3,
            then a twice repeated 3-fold CV will be performed

            ::

                default = 1
        '''

        self.cv = cv
        self.splits = splits
        self.n_repeats = n_repeats
        self._cv = _cv
        self._random_state = _random_state
        self._splits_vals = _splits_vals

    def setup(self, cv, split_vals, random_state):

        self._cv = cv
        self._splits_vals = split_vals
        self._random_state = random_state

    def get_splits(self, train_data_index):

        return self._cv.get_cv(train_data_index,
                               self.splits,
                               self.n_repeats,
                               self._splits_vals,
                               self._random_state,
                               return_index='both')


import matplotlib.pyplot as plt
import seaborn as sns

def show(self, scope):
    pass

def info(self, scope):
    pass

def plot_vars(self, scope):
    pass


def plot(self, scope, subjects=None, show_original_names=True,
         cat_types='Counts'):

    # Run check name map
    self._check_name_map()

    # Grab cols to plot
    cols = self._get_cols_from_scope(scope)

    for col in cols:
        if 'category' in self.scopes[col]:
            if len(self[col].unique()) < 50:
                self._plot_category(col, show_original_names)
            else:
                print('Skipping plot:', col, 'too many categories!')

        else:
            pass


def _plot_category(self, col, show_original_names=True):

    # Get a copy of the values
    values = self[col].copy()

    # Check for show_original_names
    if col in self.name_map:
        values = values.replace(self.name_map[col])

    # Plot
    sns.countplot(values, orient='h')



from sklearn.pipeline import Pipeline
import numpy as np
from ..helpers.VARS import ORDERED_NAMES
from .helpers import f_array


class BPtPipeline(Pipeline):

    _needs_mapping = True
    _needs_train_data_index = True

    def __init__(self, steps, memory=None, verbose=False, names=None):
        self.names = names

        super().__init__(steps=steps, memory=memory, verbose=verbose)

    @property
    def n_jobs(self):
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs):

        # Store ... in self._n_jobs
        self._n_jobs = n_jobs

        # If set here, try to propegate to all steps
        for step in self.steps:
            if hasattr(step[1], 'n_jobs'):
                setattr(step[1], 'n_jobs', n_jobs)

            # Also check for wrapper n jobs
            if hasattr(step[1], 'wrapper_n_jobs'):
                setattr(step[1], 'wrapper_n_jobs', n_jobs)

    def get_params(self, deep=True):
        params = super()._get_params('steps', deep=deep)
        return params

    def set_params(self, **kwargs):
        super()._set_params('steps', **kwargs)
        return self

    def fit(self, X, y=None, mapping=None,
            train_data_index=None, **fit_params):

        # Add mapping to fit params if already passed, e.g., in nested context
        # Or init new
        if mapping is not None:
            self._mapping = mapping.copy()
        else:
            self._mapping = {i: i for i in range(X.shape[1])}

        # Add to the fit parameters according to estimator tags
        # adding mapping + needs_train_index info
        for step in self.steps:
            name, estimator = step[0], step[1]
            if hasattr(estimator, '_needs_mapping'):
                fit_params[name + '__mapping'] = self._mapping
            if hasattr(estimator, '_needs_train_data_index'):
                fit_params[name + '__train_data_index'] = train_data_index

        # Call parent fit
        super().fit(X, y, **fit_params)
        return self

    def _get_objs_by_name(self):

        if self.names is None:
            self.names = []

        fitted_objs = [[self.__getitem__(name) for name in obj]
                       for obj in self.names]
        return fitted_objs

    def _get_ordered_objs_and_names(self, fs=True, model=False):

        fitted_objs = self._get_objs_by_name()
        ordered_objs = []
        ordered_base_names = []

        for name in ORDERED_NAMES:

            # Check if should add or not based on passed params
            add = True
            if name == 'feat_selectors' and not fs:
                add = False
            if name == 'model' and not model:
                add = False

            if add:
                ind = ORDERED_NAMES.index(name)
                ordered_objs += fitted_objs[ind]
                ordered_base_names += self.names[ind]

        return ordered_objs, ordered_base_names

    def transform_df(self, X_df, fs=True):
        '''Transform an input dataframe, keeping track of feature names'''

        # Get ordered objects as list, with or without feat selectors
        # and also the corr. base names
        ordered_objs, ordered_base_names =\
            self._get_ordered_objs_and_names(fs=fs, model=False)

        # Run all of the transformations
        for obj, base_name in zip(ordered_objs, ordered_base_names):
            X_df = obj.transform_df(X_df, base_name=base_name)

        return X_df

    def inverse_transform_FIs(self, fis, feat_names):

        # Make compat w/ subjects x feats
        if len(fis.shape) == 1:
            fis = np.expand_dims(fis, axis=0)

        # To inverse transform FIs, we are only concerned with feat_selectors
        # transformers, and loaders
        fitted_objs = self._get_objs_by_name()

        # Feat selectors
        fs_ind = ORDERED_NAMES.index('feat_selectors')
        for feat_selector in fitted_objs[fs_ind][::-1]:
            fis = feat_selector.inverse_transform(fis)

        # Transformers
        trans_ind = ORDERED_NAMES.index('transformers')
        for transformer, name in zip(fitted_objs[trans_ind][::-1],
                                     self.names[trans_ind][::-1]):
            fis = transformer.inverse_transform(fis, name=name)

        # Loaders - special case
        inversed_loaders = {}
        l_ind = ORDERED_NAMES.index('loaders')
        for loader, name in zip(fitted_objs[l_ind][::-1],
                                self.names[l_ind][::-1]):
            fis, inverse_X = loader.inverse_transform(fis, name=name)
            inversed_loaders.update(inverse_X)

        # Make the final feat_importances dict
        feat_imp_dict = {}
        for i in range(len(feat_names)):
            if i in inversed_loaders:
                feat_imp_dict[feat_names[i]] = inversed_loaders[i]
            else:
                feat_imp_dict[feat_names[i]] = fis[:, i]

        return feat_imp_dict



class Select(list):
    '''The Select object is an BPt specific Input Wrapper designed
    to allow hyper-parameter searches to include not
    just choice of hyper-parameter,
    but also choosing between objects
    (as well as their relevant distributions).

    Select is used to cast lists of base :class:`Model_Pipeline`
    pieces as different options.
    Consider a simple example, for specifying a selection
    between two different :class:`Models<Model>`

    ::

        model = Select([Model('linear'), Model('random forest')])

    In this example, the model passed to :class:`Model_Pipeline`
    becomes a meta object
    for selecting between the two base models.
    Note: this does require a :class:`Param_Search`
    object be passed to :class:`Model_Pipeline`.
    Notably as well, if further param distributions
    are defined within say the `Model('random forest')`,
    those will still be optimized, allowing for
    potentially even a hyper-parameter search to
    select hyper-parameter distribution...
    (i.e., if both select options had the same base model obj,
    but only differed in the passed hyper-param distributions) if one were
    so inclined...

    Other notable features of Select are, you are not limited to
    passing only two options,
    you can pass an arbitrary number... you can even,
    and I'm not even sure I want to tell you this...
    pass nested calls to Select... i.e., one of the
    Select options could be another Select, with say
    another Select...

    Lastly, explcitly note that Select is not restricted for use with Models,
    it can be used on any of
    the base class:`Model_Pipeline` piece params
    (i.e., every param but param_search, feat_importances and cache...).
    '''
    input_type = 'select'

    def __repr__(self):
        return 'Select(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()


def is_select(obj):

    try:
        if obj.input_type == 'select':
            return True
        return False

    except AttributeError:
        return False


class Duplicate(list):
    '''The Duplicate object is an BPt specific Input wrapper.
    It is designed to be cast on a list of valid scope parameters, e.g., 

    ::

        scope = Duplicate(['float', 'cat'])

    Such that the corresponding pipeline piece will be duplicated for every
    entry within Duplicate. In this case, two copies of the base object will be
    made, where both have the same remaining non-scope params
    (i.e., obj, params, extra_params),
    but one will have a scope of 'float' and the other 'cat'.

    Consider the following exentended example, where loaders is being specified
    when creating an instance of :class:`Model_Pipeline`:

    ::

        loaders = Loader(obj='identity', scope=Duplicate(['float', 'cat']))

    Is transformed in post processing / equivalent to

    ::

        loaders = [Loader(obj='identity', scope='float'),
                   Loader(obj='identity', scope='cat')]

    '''

    input_type = 'duplicate'

    def __repr__(self):
        return 'Duplicate(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()


def is_duplicate(obj):

    try:
        if obj.input_type == 'duplicate':
            return True
        return False

    except AttributeError:
        return False


class Pipe(list):
    '''The Pipe object is an BPt specific Input wrapper, designed
    for now to work specifically within :class:`Loader`.
    Because loader
    objects within BPt are designed to work on single files at a time,
    and further are resitricted in that they must go
    directly from some arbitrary
    file, shape and charteristics to outputted
    as a valid 2D (# Subects X # Features) array,
    it restricts potential sequential compositions.
    Pipe offers some utilty towards
    building sequential compositions.

    For example, say one had saved 4D neuroimaging fMRI timeseries,
    and they wanted
    to first employ a loader to extract timeseries by ROI
    (with say hyper-parameters defined to select which ROI to use),
    but then wanted to use
    another loader to convert the timeseries ROIs to a correlation matrix,
    and only then pass
    along the output as 1D features per subject.
    In this case, the Pipe wrapper is a greate canidate!

    Specifically, the pipe wrapper works at the level of defining a
    specific Loader, where basicially
    you are requesting that the loader you want to use be a
    Pipeline of a few different loader options,
    where the loader options are ones compatible in
    passing input to each other, e.g., the output from
    fit_transform as called on the ROI extractor is valid input
    to fit_transform of the Timeseries creator,
    and lastly the output from fit_transform of the
    Timeseries creator valid 1D feature array per subjects output.

    Consider the example in code below, where we
    assume that 'rois' is the ROI extractor,
    and 'timeseries' is the correlation matrix
    creator object (where these could be can valid loader str, or
    custom user passed objects)

    ::

        loader = Loader(obj = Pipe(['rois', 'timeseries']))

    We only passed arguments for obj above, but in our toy example
    as initially described we wanted to
    further define parameters for a parameter search across both objects.
    See below for what different options
    for passing corresponding parameter distributions are:

    ::

        # Options loader1 and loader2 tell it explicitly no params

        # Special case, if just default params = 0, will convert to 2nd case
        loader1 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = 0)

        # You can specify just a matching list
        loader2 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [0, 0])

        # Option 3 assumes that there are pre-defined valid class param dists
        # for each of the base objects
        loader3 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [1, 1])

        # Option 4 lets set params for the 'rois' object, w/ custom param dists
        loader4 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [{'some custom param dist'}, 0])

    Note that still only one scope may be passed, and that scope will
    define the scope of the new combined loader.
    Also note that if extra_params is passed, the same extra_params will
    be passed when creating both individual objects.
    Where extra params behavior is to add its contents, only when the
    name of that param appears in the base classes init, s.t.
    there could exist a case where, if both 'rois' and 'timeseries'
    base objects had a parameter with the same name, passing a
    value for that name in extra params would update them
    both with the passed value.
    '''

    input_type = 'pipe'

    def __repr__(self):
        return 'Pipe(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()


def is_pipe(obj):

    try:
        if obj.input_type == 'pipe':
            return True
        return False

    except AttributeError:
        return False


class Value_Subset():
    ''' Value_Subset is special wrapper class for BPt designed to work with
    :ref:`Subjects` style input. As seen in :class:`Param_Search`,
    or to the `train_subjects` or `test_subjects`
    params in :func:`Evaluate <BPt.BPt_ML.Evaluate>`
    and :func:`Test <BPt.BPt_ML.Test>`.

     This wrapper can be used as follows, just specify an object as

     ::

        Value_Subset(name, value)

    Where name is the name of a loaded Strat column / feature,
    and value is the subset of values from that column to select subjects by.
    E.g., if you wanted to select just subjects of a specific sex,
    and assuming a variable was
    loaded in Strat (See :func:`Load_Strat <BPt.BPt_ML.Load_Strat>`)
    you could pass:

    ::

        subjects = Value_Subset('sex', 0)

    Which would specify only subjects with 'sex' equal to 0.
    You may also pass a list-like set of multiple columns to the name param.
    In this case, the overlap
    across all passed names will be computed, for example:

    ::

        subjects = Value_Subset(['sex', 'race'], 0)

    Where 'race' is another valid loaded Strat, would select
    only subjects with a value of 0 in the computed unique
    overlap across 'sex' and 'race'.

    Note it might be hard to tell what a value of 0 actually means,
    especially when you compose
    across multiple variables. With that in mind, as long as verbose
    is set to True, upon computation
    of the subset of subjects a message with be printed indicating
    what the passed value corresponds to
    in all of the combined variables, e.g., in the example above
    you would get the print out 'sex' = 0, 'race' = 0.
    '''
    input_type = 'value_subset'

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return 'Value_Subset(name=' + str(self.name) + ', value=' + \
          str(self.value) + ')'

    def __str__(self):
        return self.__repr__()


def is_value_subset(obj):

    try:
        if obj.input_type == 'value_subset':
            return True
        return False

    except AttributeError:
        return False


class Values_Subset():
    '''Value_Subsets is special wrapper class for BPt designed to work with
    :ref:`Subjects` style input.

    This wrapper is very similar to :class:`Value_Subject`, and will actually
    function the same in the case that one value for name and one value
    for values is selected, e.g. the below are equivilent.

    ::

        subjects = Value_Subset(name='sex', value=0)
        subjects = Values_Subset(name='sex', values=0)

    That said, where Value_Subset, allows passing multiple values for name,
    but only allows one value for value, Values_Subset only allows one
    value for name, and multiple values for values.

    Values_Subset therefore lets you select the subset of subjects via
    one or more values in a loaded Strat variable. E.g.,

    ::

        subjects = Values_Subset(name='site', values=[0,1,5])

    Would select the subset of subjects from sites 0, 1 and 5.
    '''

    input_type = 'values_subset'

    def __init__(self, name, values):
        self.name = name
        self.values = values

        if not isinstance(self.name, str):
            raise ValueError('name must be a string!')

    def __repr__(self):
        return 'Values_Subset(name=' + str(self.name) + ', values=' + \
          str(self.values) + ')'

    def __str__(self):
        return self.__repr__()


def is_values_subset(obj):

    try:
        if obj.input_type == 'values_subset':
            return True
        return False

    except AttributeError:
        return False


def is_special(obj):
    return hasattr(obj, 'input_type')



"""
Feature_Selectors.py
====================================
File with different Feature Selectors
"""
from ..helpers.ML_Helpers import (get_possible_init_params,
                                  get_obj_and_params)
from sklearn.feature_selection import SelectPercentile, VarianceThreshold
from ..extensions.Feat_Selectors import RFE_Wrapper, FeatureSelector
import numpy as np
from numpy.random import RandomState
import nevergrad as ng

from sklearn.base import BaseEstimator, clone
from sklearn.feature_selection._base import SelectorMixin
from ..helpers.ML_Helpers import proc_mapping, update_mapping
from .ScopeObjs import ScopeTransformer


class BPtFeatureSelector(ScopeTransformer, SelectorMixin):

    def fit(self, X, y=None, mapping=None,
            train_data_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit
        super().fit(X, y=y, mapping=mapping,
                    train_data_index=train_data_index,
                    **fit_params)

        # Need to pass along the correct mapping
        # overwrite existing out mapping
        self.out_mapping_ = {}

        # This is the calculated support from the base estimator
        support = self.estimator_.get_support()

        # First half is for updating the index within scope
        cnt = 0
        for i, ind in enumerate(self.inds_):

            # If kept by feat selection, add, otherwise set to None
            if support[i]:
                self.out_mapping_[ind] = cnt
                cnt += 1
            else:
                self.out_mapping_[ind] = None

        # Next, need to update the mapping for the remaining wrapper inds
        # essentially setting them where the cnt left off, then sequentially
        for rem_ind in range(len(self.rest_inds_)):
            self.out_mapping_[self.rest_inds_[rem_ind]] = cnt
            cnt += 1

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

        return self

    def _proc_new_names(self, feat_names, base_name=None):

        # Get base new names from parent class
        new_names = super()._proc_new_names(feat_names)

        # This feat mask corresponds to the already transformed feats
        feat_mask = self._get_support_mask()

        # Apply the computed mask to get the actually selected features
        return_names = np.array(new_names)[feat_mask]

        return list(return_names)

    def _get_support_mask(self):

        # Create full support as base support + True's for all rest inds
        # i.e., those features originally out of scope
        base_support = self.estimator_.get_support()
        rest_support = np.ones(len(self.rest_inds_), dtype='bool')
        support = np.concatenate([base_support, rest_support])

        return support


AVALIABLE = {
        'binary': {
                'univariate selection c':
                'univariate selection c',
                'univariate selection':
                'univariate selection c',
                'rfe': 'rfe',
                'variance threshold': 'variance threshold',
                'selector': 'selector',
        },
        'regression': {
                'univariate selection r':
                'univariate selection r',
                'univariate selection':
                'univariate selection r',
                'rfe': 'rfe',
                'variance threshold': 'variance threshold',
                'selector': 'selector',
        },
}

AVALIABLE['categorical'] = AVALIABLE['binary'].copy()

SELECTORS = {
    'univariate selection r': (SelectPercentile,
                               ['base univar fs regression',
                                'univar fs regression dist',
                                'univar fs regression dist2']),

    'univariate selection c': (SelectPercentile,
                               ['base univar fs classifier',
                                'univar fs classifier dist',
                                'univar fs classifier dist2']),

    'rfe': (RFE_Wrapper, ['base rfe', 'rfe num feats dist']),

    'variance threshold': (VarianceThreshold, ['default']),

    'selector': (FeatureSelector, ['random', 'searchable']),
}


def get_special_selector(feat_selector, feat_selector_params, random_state,
                         num_feat_keys):

    # Init feat selector with mask of random feats
    if random_state is None:
        r_state = RandomState(np.random.randint(1000))
    elif isinstance(random_state, int):
        r_state = RandomState(random_state)
    else:
        r_state = random_state

    init_mask = r_state.random(num_feat_keys)
    feat_selector = feat_selector(mask=init_mask)

    # Figure out param passed
    if 'selector__mask' in feat_selector_params:
        p_name = 'selector__mask'

        # If set to searchable, set to searchable...
        if feat_selector_params[p_name] == 'sets as hyperparameters':

            feat_array = ng.p.Array(init=[.5 for i in range(num_feat_keys)])
            feat_array.set_mutation(sigma=1/6).set_bounds(lower=0, upper=1)
            feat_selector_params[p_name] = feat_array

        elif feat_selector_params[p_name] == 'sets as random features':
            del feat_selector_params[p_name]

    return feat_selector, feat_selector_params


def get_feat_selector_and_params(feat_selector_str, extra_params, params,
                                 random_state, num_feat_keys):
    '''Returns a scaler based on proced str indicator input,

    Parameters
    ----------
    feat_selector_str : str
        `feat_selector_str` refers to the type of feature selection
        to apply to the saved data during model evaluation.

    extra_params : dict
        Any extra params being passed.
        These can be supplied by creating another dict within extra_params.
        E.g., extra_params[method] = {'method param' : new_value}
        Where method param is a valid argument for that method,
        and method in this case is the str indicator.

    params : int
        The index of the params to use.

    Returns
    ----------
    feature_selector
        A feature selector object with fit and transform methods.

    dict
        The params for this feature selector
    '''

    feat_selector, extra_feat_selector_params, feat_selector_params =\
        get_obj_and_params(feat_selector_str, SELECTORS, extra_params, params)

    # Special behavior for selector...
    if feat_selector_str == 'selector':

        feat_selector, feat_selector_params =\
            get_special_selector(feat_selector, feat_selector_params,
                                 random_state, num_feat_keys)

        return feat_selector, feat_selector_params

    else:
        # Need to check for estimator, as RFE needs a default param for est.
        # Set as placeholder None if passed
        possible_params = get_possible_init_params(feat_selector)
        if 'estimator' in possible_params:
            extra_feat_selector_params['estimator'] = None

        return (feat_selector(**extra_feat_selector_params),
                feat_selector_params)



import pandas as pd

import numpy as np
import time

from ..helpers.ML_Helpers import conv_to_list
from .Feat_Importances import get_feat_importances_and_params
from .Scorers import process_scorers
from copy import deepcopy
from os.path import dirname, abspath, exists
from sklearn.base import clone


class Evaluator():
    '''Helper class for handling all of the different parameters involved in
    model training, scaling, handling different datatypes ect...
    '''

    def __init__(self, model, problem_spec, cv, all_keys,
                 feat_importances, return_raw_preds, return_models,
                 verbosity, base_dtype, _print=print):

        # Save passed params
        self.model = model
        self.ps = problem_spec
        self.cv = cv
        self.all_keys = all_keys
        self.return_raw_preds = return_raw_preds
        self.return_models = return_models
        self.progress_bar = verbosity['progress_bar']
        self.compute_train_score = verbosity['compute_train_score']
        self.progress_loc = verbosity['progress_loc']
        self.base_dtype = base_dtype
        self._print = _print
        self.models = []

        # Default params
        self._set_default_params()

        # Set / proc scorers
        self.scorer_strs, self.scorers, _ =\
            process_scorers(deepcopy(self.ps.scorer), self.ps.problem_type)

        # Set / proc feat importance info
        self.feat_importances =\
            self._process_feat_importances(feat_importances)

    def _set_default_params(self):

        self.n_splits_ = None

        self.flags = {'linear': False,
                      'tree': False}

    def _process_feat_importances(self, feat_importances):

        # Grab feat_importance from spec as a list
        feat_importances = conv_to_list(feat_importances)

        if feat_importances is not None:

            scorers = [fi.scorer for fi in feat_importances]
            scorers =\
                [process_scorers(m, self.ps.problem_type)[2] for m in scorers]

            feat_importances =\
                [get_feat_importances_and_params(fi, self.ps.problem_type,
                                                 self.ps.n_jobs, scorer)
                    for fi, scorer in zip(feat_importances, scorers)]

            return feat_importances
        return []

    def _get_subjects_overlap(self, subjects):
        '''Computer overlapping subjects with self.ps._final_subjects'''

        if self.ps._final_subjects is None:
            overlap = set(subjects)
        else:
            overlap = self.ps._final_subjects.intersection(set(subjects))

        return np.array(list(overlap))

    def _get_results(self):

        results = {}

        # If raw_preds off, will just return None
        results['raw_preds'] = self.raw_preds_df

        # If no feature importances will just be empty list
        results['FIs'] = self.feat_importances

        # If return models off, will just be empty list
        results['models'] = self.models

        return results

    def Evaluate(self, data, train_subjects, splits,
                 n_repeats, splits_vals, only_fold=None):
        '''Method to perform a full evaluation
        on a provided model type and training subjects, according to
        class set parameters.

        Parameters
        ----------
        data : pandas DataFrame
            BPt formatted, with both training and testing data.

        train_subjects : array-like
            An array or pandas Index of the train subjects should be passed.

        Returns
        ----------
        array-like of array-like
            numpy array of numpy arrays,
            where each internal array contains the raw scores as computed for
            all passed in scorers, computed for each fold within
            each repeat.
            e.g., array will have a length of `n_repeats` * n_splits
            (num folds) and each internal array will have the same length
            as the number of scorers.
        '''

        # Set train_subjects according to self.ps._final_subjects
        train_subjects = self._get_subjects_overlap(train_subjects)

        # Init raw_preds_df
        self._init_raw_preds_df(train_subjects)

        # Setup the desired eval splits
        subject_splits =\
            self._get_eval_splits(train_subjects, splits,
                                  n_repeats, splits_vals,
                                  only_fold=only_fold)

        all_train_scores, all_scores = [], []
        fold_ind = 0

        # Init progress bar if any
        if self.progress_bar is not None:
            repeats_bar = self.progress_bar(total=n_repeats,
                                            desc='Repeats')

            folds_bar = self.progress_bar(total=self.n_splits_,
                                          desc='Folds')

        # Init progress loc if any
        if self.progress_loc is not None:
            with open(self.progress_loc, 'w') as f:
                f.write(str(n_repeats) + ',' + str(self.n_splits_))
                f.write('\n')

        self.n_test_per_fold = []

        # For each split with the repeated K-fold
        for train_subjects, test_subjects in subject_splits:

            self.n_test_per_fold.append(len(test_subjects))

            # Fold name verbosity
            repeat = str((fold_ind // self.n_splits_) + 1)
            fold = str((fold_ind % self.n_splits_) + 1)
            self._print(level='name')
            self._print('Repeat: ', repeat, '/', n_repeats, ' Fold: ',
                        fold, '/', self.n_splits_, sep='', level='name')

            if self.progress_bar is not None:
                repeats_bar.n = int(repeat) - 1
                repeats_bar.refresh()

                folds_bar.n = int(fold) - 1
                folds_bar.refresh()

            # Run actual code for this evaluate fold
            start_time = time.time()
            train_scores, scores = self.Test(data, train_subjects,
                                             test_subjects, fold_ind)

            # Time by fold verbosity
            elapsed_time = time.time() - start_time
            time_str = time.strftime("%H:%M:%S", time.gmtime(elapsed_time))
            self._print('Time Elapsed:', time_str, level='time')

            # Score by fold verbosity
            if self.compute_train_score:
                for i in range(len(self.scorer_strs)):
                    self._print('train ', self.scorer_strs[i], ': ',
                                train_scores[i], sep='', level='score')

            for i in range(len(self.scorer_strs)):
                self._print('val ', self.scorer_strs[i], ': ',
                            scores[i], sep='', level='score')

            # If progress loc
            if self.progress_loc is not None:

                if not exists(dirname(abspath(self.progress_loc))):
                    raise SystemExit('Folder where progress is stored '
                                     ' was removed!')

                with open(self.progress_loc, 'a') as f:
                    f.write('fold\n')

            all_train_scores.append(train_scores)
            all_scores.append(scores)
            fold_ind += 1

        if self.progress_bar is not None:
            repeats_bar.n = n_repeats
            repeats_bar.refresh()
            repeats_bar.close()

            folds_bar.n = self.n_splits_
            folds_bar.refresh()
            folds_bar.close()

        # If any local feat importances
        for feat_imp in self.feat_importances:
            feat_imp.set_final_local()

        # self.micro_scores = self._compute_micro_scores()

        results = self._get_results()
        return (np.array(all_train_scores), np.array(all_scores), results)

    def _get_eval_splits(self, train_subjects, splits,
                         n_repeats, splits_vals, only_fold=None):

        subject_splits = self.cv.get_cv(train_subjects, splits, n_repeats,
                                        splits_vals, self.ps.random_state,
                                        return_index=False)

        # Want to save n_splits by type
        if splits_vals is not None:
            self.n_splits_ = self.cv.get_num_groups(train_subjects,
                                                    splits_vals)

        elif isinstance(splits, int):
            self.n_splits_ = splits

        else:
            self.n_splits_ = 1

        # Check for passed only_fold
        if only_fold is not None:

            subject_splits = [subject_splits[only_fold]]
            self.n_splits_ = 1

        return subject_splits

    def Test(self, data, train_subjects, test_subjects, fold_ind='test'):
        '''Method to test given input data, training a model on train_subjects
        and testing the model on test_subjects.

        Parameters
        ----------
        data : pandas DataFrame
            BPt formatted, with both training and testing data.

        train_subjects : array-like
            An array or pandas Index of train subjects should be passed.

        test_subjects : array-like
            An array or pandas Index of test subjects should be passed.

        Returns
        ----------
        array-like
            A numpy array of scores as determined by the passed
            metric/scorer(s) on the provided testing set.
        '''

        # Reset progress loc if Test
        if fold_ind == 'test':
            if self.progress_loc is not None:
                if not exists(dirname(abspath(self.progress_loc))):
                    raise SystemExit('Folder where progress is stored '
                                     ' was removed!')

                with open(self.progress_loc, 'w') as f:
                    f.write('test\n')

        # Ensure train and test subjects are just the requested overlap
        train_subjects = self._get_subjects_overlap(train_subjects)

        # Remove any train subjects with NaN targets
        train_targets = data.loc[train_subjects, self.ps.target]
        if pd.isna(train_targets).any():
            valid = train_targets[~pd.isna(train_targets)].index
            train_subjects = self._get_subjects_overlap(valid)

        # For test subjects set to overlap
        all_test_subjects = self._get_subjects_overlap(test_subjects)

        # Get another version of the test_subjects w/o NaN
        test_targets = data.loc[all_test_subjects, self.ps.target]
        if pd.isna(test_targets).any():
            valid = test_targets[~pd.isna(test_targets)].index
            test_subjects = self._get_subjects_overlap(valid)
        else:
            test_subjects = all_test_subjects

        # Ensure data being used is just the selected col / feats
        data = data[self.all_keys]

        # Init raw_preds_df
        if fold_ind == 'test':

            # For raw preds df, keep NaNs, so use all_test_subjects
            if self.compute_train_score:
                self._init_raw_preds_df(np.concatenate([train_subjects,
                                                        all_test_subjects]))
            else:
                self._init_raw_preds_df(all_test_subjects)

        # Assume the train_subjects here are final
        train_data = data.loc[train_subjects]

        self._print('Train shape:', train_data.shape, level='size')
        self._print('Val/Test shape:', data.loc[test_subjects].shape,
                    level='size')
        if len(test_subjects) != len(all_test_subjects):
            self._print('Making predictions for additional target NaN '
                        'subjects:',
                        len(all_test_subjects) - len(test_subjects),
                        level='size')

        # Train the model(s)
        self._train_model(train_data)

        # Proc the different feat importances,
        # Pass only test subjects w/o missing targets here
        self._proc_feat_importance(train_data, data.loc[test_subjects],
                                   fold_ind)

        # Get the scores
        if self.compute_train_score:
            train_scores = self._get_scores(train_data, 'train_', fold_ind)
        else:
            train_scores = 0

        # Pass test_data w/ Nans to get_scores, in order to
        # still record predictions for targets w/ a missing
        # ground truth.
        scores = self._get_scores(data.loc[all_test_subjects], '', fold_ind)

        # Return differently based on if test
        if fold_ind == 'test':
            results = self._get_results()
            return (train_scores, scores, results)

        return train_scores, scores

    def _get_base_fitted_pipeline(self):

        if hasattr(self.model_, 'name') and self.model_.name == 'search':
            return self.model_.best_estimator_

        return self.model_

    def _get_base_fitted_model(self):

        base_pipeline = self._get_base_fitted_pipeline()
        last_name = base_pipeline.steps[-1][0]
        base_model = base_pipeline[last_name]

        return base_model

    def _set_model_flags(self):

        base_model = self._get_base_fitted_model()

        try:
            base_model.coef_
            self.flags['linear'] = True
        except AttributeError:
            pass

        try:
            base_model.feature_importances_
            self.flags['tree'] = True
        except AttributeError:
            pass

    def _proc_feat_importance(self, train_data, test_data, fold_ind):

        # Ensure model flags are set / there are feat importances to proc
        if len(self.feat_importances) > 0:
            self._set_model_flags()
        else:
            return

        # Process each feat importance
        for feat_imp in self.feat_importances:

            split = feat_imp.split

            # Init global feature df
            if fold_ind == 0 or fold_ind == 'test':

                X, y = self._proc_X_test(train_data, fs=False)
                feat_imp.init_global(X, y)

            # Local init - Test
            if fold_ind == 'test':

                if split == 'test':
                    X, y = self._proc_X_test(test_data, fs=False)

                elif split == 'train':
                    X, y = self._proc_X_test(train_data, fs=False)

                elif split == 'all':
                    X, y =\
                        self._proc_X_test(pd.concat([train_data, test_data]),
                                          fs=False)

                # Error if here
                else:
                    X, y = None, None

                feat_imp.init_local(X, y, test=True, n_splits=None)

            # Local init - Evaluate
            elif fold_ind % self.n_splits_ == 0:

                X, y = self._proc_X_test(pd.concat([train_data, test_data]),
                                         fs=False)
                feat_imp.init_local(X, y, n_splits=self.n_splits_)

            # Error if reaches here
            else:
                X, y = None, None

            self._print('Calculate', feat_imp.name, 'feat importances',
                        level='name')

            # Get base fitted model
            base_model = self._get_base_fitted_model()

            # Optionally proc train, though train is always train
            if feat_imp.get_data_needed_flags(self.flags):
                X_train = self._proc_X_train(train_data)
            else:
                X_train = None

            # Test depends on scope
            if split == 'test':
                test = test_data
            elif split == 'train':
                test = train_data
            elif split == 'all':
                test = pd.concat([train_data, test_data])

            # Always proc test.
            X_test, y_test = self._proc_X_test(test)

            try:
                fold = fold_ind % self.n_splits_
            except TypeError:
                fold = 'test'

            # Process the feature importance, provide all needed
            fis =\
                feat_imp.proc_importances(base_model, X_test, y_test=y_test,
                                          X_train=X_train, fold=fold,
                                          random_state=self.ps.random_state)

            # Grab the names of all input features
            feat_names = list(train_data)
            feat_names.remove(self.ps.target)

            # Inverse transform FIs back to original feat_space is requested
            self._inverse_transform_FIs(feat_imp, fis, feat_names)

            # For local, need an intermediate average, move df to dfs
            if isinstance(fold_ind, int):
                if fold_ind % self.n_splits_ == self.n_splits_-1:
                    feat_imp.proc_local()

    def _inverse_transform_FIs(self, feat_imp, fis, feat_names):

        global_fi, local_fi = fis
        pipeline = self._get_base_fitted_pipeline()

        if feat_imp.inverse_global and global_fi is not None:
            feat_imp.inverse_global_fis.append(
                pipeline.inverse_transform_FIs(global_fi, feat_names))

        if feat_imp.inverse_local and local_fi is not None:
            feat_imp.inverse_local_fis.append(
                pipeline.inverse_transform_FIs(local_fi, feat_names))

        feat_imp.warning = True
        return

    def _get_X_y(self, data, X_as_df=False, copy=False):
        '''Helper method to get X,y data from BPt formatted df.

        Parameters
        ----------
        data : pandas DataFrame
            BPt formatted.

        X_as_df : bool, optional
            If True, return X as a pd DataFrame,
            otherwise, return as a numpy array

            (default = False)

        copy : bool, optional
            If True, return a copy of X

            (default = False)

        Returns
        ----------
        array-like
            X data for ML
        array-like
            y target for ML
        '''

        if copy:
            X = data.drop(self.ps.target, axis=1).copy()
            y = data[self.ps.target].copy()
        else:
            X = data.drop(self.ps.target, axis=1)
            y = data[self.ps.target]

        if not X_as_df:
            X = np.array(X).astype(self.base_dtype)

        y = np.array(y).astype('float64')

        return X, y

    def _train_model(self, train_data):
        '''Helper method to train a models given
        a str indicator and training data.

        Parameters
        ----------
        train_data : pandas DataFrame
            BPt formatted, training data.

        Returns
        ----------
        sklearn api compatible model object
            The trained model.
        '''

        # Data, score split
        X, y = self._get_X_y(train_data)

        # Fit the model
        self.model_ = clone(self.model)
        self.model_.fit(X, y, train_data_index=train_data.index)

        # If return models, save model
        if self.return_models:
            self.models.append(self.model_)

        # If a search object, try to print best score
        try:
            self._show_best_score()
        except KeyError:
            self._print('Error printing best score',
                        level='cv_score')

        # If a search object, show the best params
        try:
            self._show_best_params()
        except KeyError:
            self._print('Error printing best params',
                        level='params')

    def _show_best_score(self):

        try:
            name = self.model_.name
            if name != 'search':
                return None
        except AttributeError:
            return None

        self._print('Best Params Score:',
                    self.model_.best_score_, level='cv_score')

    def _show_best_params(self):

        try:
            name = self.model_.name
            if name != 'search':
                return None
        except AttributeError:
            return None

        self._print('Params Selected by Pipeline:', level='params')
        self._print(self.model_.best_params_, level='params')

        return None

    def _get_all_params(self, params, all_ps, to_show):

        if any(['__select' in p for p in params]):

            all_params = {}
            for p in params:
                ud = params[p]

                if '__select' in p:
                    base_name = '__'.join(p.split('__')[:-1])
                    to_use = all_ps[base_name + '__to_use']
                    to_show.append(base_name + '__selected: ' + str(to_use))
                    extra_ps = ud.choices[to_use]._content
                    extra_ps =\
                        {base_name + '__' + e: extra_ps[e] for e in extra_ps}

                    all_params.update(extra_ps)
                else:
                    all_params[p] = params[p]

            # Recursively call
            return self._get_all_params(all_params, all_ps, to_show)

        return params, to_show

    def _get_scores(self, test_data, eval_type, fold_ind):
        '''Helper method to get the scores of
        the trained model saved in the class on input test data.
        For all metrics/scorers.

        Parameters
        ----------
        test_data : pandas DataFrame
            BPt formatted test data.

        eval_type : {'train_', ''}

        fold_ind : int or 'test'

        Returns
        ----------
        float
            The score of the trained model on the given test data.
        '''

        # Data, score split
        X_test, y_test = self._get_X_y(test_data)

        # For book-keeping set num y classes
        self._set_classes(y_test)

        # Add raw preds to raw_preds_df
        self._add_raw_preds(X_test, y_test,
                            test_data.index, eval_type,
                            fold_ind)

        # Only compute scores on Non-Nan y
        non_nan_mask = ~np.isnan(y_test)

        # Get the scores
        scores = [scorer(self.model_,
                         X_test[non_nan_mask],
                         y_test[non_nan_mask])
                  for scorer in self.scorers]

        return np.array(scores)

    def _set_classes(self, y_test):

        # Get non-nan classes
        self.classes = np.unique(y_test[~np.isnan(y_test)])

        # Catch case where there is only one class present in y_test
        # Assume in this case that it should be binary, 0 and 1
        if len(self.classes) == 1:
            self.classes = np.array([0, 1])

    def _add_raw_preds(self, X_test, y_test, subjects, eval_type, fold_ind):

        # If return_raw_preds set to false, skip
        if not self.return_raw_preds:
            return

        if fold_ind == 'test':
            fold = 'test'
            repeat = ''
        else:
            fold = str((fold_ind % self.n_splits_) + 1)
            repeat = str((fold_ind // self.n_splits_) + 1)

        try:
            raw_prob_preds = self.model_.predict_proba(X_test)
            pred_col = eval_type + repeat + '_prob'

            if len(np.shape(raw_prob_preds)) == 3:

                for i in range(len(raw_prob_preds)):
                    p_col = pred_col + '_class_' + str(self.classes[i])
                    class_preds = [val[1] for val in raw_prob_preds[i]]
                    self.raw_preds_df.loc[subjects, p_col] = class_preds

            elif len(np.shape(raw_prob_preds)) == 2:

                for i in range(np.shape(raw_prob_preds)[1]):
                    p_col = pred_col + '_class_' + str(self.classes[i])
                    class_preds = raw_prob_preds[:, i]
                    self.raw_preds_df.loc[subjects, p_col] = class_preds

            else:
                self.raw_preds_df.loc[subjects, pred_col] = raw_prob_preds

        except AttributeError:
            pass

        raw_preds = self.model_.predict(X_test)
        pred_col = eval_type + repeat

        if len(np.shape(raw_preds)) == 2:
            for i in range(np.shape(raw_preds)[1]):
                p_col = pred_col + '_class_' + str(self.classes[i])
                class_preds = raw_preds[:, i]
                self.raw_preds_df.loc[subjects, p_col] = class_preds

        else:
            self.raw_preds_df.loc[subjects, pred_col] = raw_preds

        self.raw_preds_df.loc[subjects, pred_col + '_fold'] = fold

        # Make copy of true values
        if len(np.shape(y_test)) > 1:
            for i in range(len(self.ps.target)):
                self.raw_preds_df.loc[subjects, self.ps.target[i]] =\
                    y_test[:, i]

        elif isinstance(self.ps.target, list):
            t_base_key = '_'.join(self.ps.target[0].split('_')[:-1])
            self.raw_preds_df.loc[subjects, 'multiclass_' + t_base_key] =\
                y_test

        else:
            self.raw_preds_df.loc[subjects, self.ps.target] = y_test

    def _init_raw_preds_df(self, subjects):

        if self.return_raw_preds:
            self.raw_preds_df = pd.DataFrame(index=subjects)
        else:
            self.raw_preds_df = None

    def _proc_X_test(self, test_data, fs=True):

        # Grab the test data, X as df + copy
        X_test, y_test = self._get_X_y(test_data, X_as_df=True, copy=True)

        # Get the base pipeline
        pipeline = self._get_base_fitted_pipeline()

        # Transform X as df
        X_trans_df = pipeline.transform_df(X_test, fs=fs)

        return X_trans_df, y_test

    def _proc_X_train(self, train_data):

        # Get X train
        X_train, _ = self._get_X_y(train_data, copy=True)

        # Get the base pipeline
        pipeline = self._get_base_fitted_pipeline()

        # Transform X
        return pipeline.transform(X_train)



from joblib import dump, load, Parallel, delayed
import os
import copy
import numpy as np


def get_feat_importances(model_loc, scorer, inds, X, y, n_perm):

    results = {}
    model = load(model_loc)

    X_copy = X.copy()

    for ind in inds:

        original = X_copy[:, ind].copy()
        importances = []

        for perm in range(n_perm):
            X_copy[:, ind] = np.random.permutation(X_copy[:, ind])
            importances.append(scorer(model, X_copy, y))

        results[ind] = np.mean(importances)
        X_copy[:, ind] = original

    return results


class Perm_Feat_Importance():

    def __init__(self, n_perm=1, n_jobs=1, temp_dr=''):

        self.n_perm = n_perm
        self.n_jobs = n_jobs
        self.temp_dr = temp_dr

    def get_chunks(self):

        per_chunk = self.X.shape[1] // self.n_jobs
        chunks = [list(range(i * per_chunk, (i+1) * per_chunk))
                  for i in range(self.n_jobs)]

        last = chunks[-1][-1]
        chunks[-1] += list(range(last+1, self.X.shape[-1]))
        return chunks

    def compute(self, model, scorer, X, y):

        self.X = X
        self.y = y

        baseline_score = scorer(model, self.X, self.y)

        if self.n_jobs == 1:
            scores = []
            for ind in range(X.shape[1]):
                scores.append(self.get_feat_importance(ind, scorer, model))
        else:

            changed = None

            # Ensure models n_jobs set to 1
            try:
                original = model.n_jobs
                model.n_jobs = 1
                changed = original

            except AttributeError:
                pass

            rn = str(np.random.random())
            model_loc = os.path.join(self.temp_dr, 'temp' + rn + '.joblib')
            dump(model, model_loc)
            model_locs = [model_loc for i in range(self.n_jobs)]

            scorers = [scorer for i in range(self.n_jobs)]
            inds = self.get_chunks()
            Xs = [X for i in range(self.n_jobs)]
            ys = [y for i in range(self.n_jobs)]
            n_perms = [self.n_perm for i in range(self.n_jobs)]

            imp_dicts =\
                Parallel(n_jobs=self.n_jobs)(delayed(get_feat_importances)
                                                    (ml, s, i, x, y, n) for
                                             ml, s, i, x, y, n in zip(
                                                 model_locs,
                                                 scorers,
                                                 inds, Xs, ys,
                                                 n_perms))

            scores = [0 for i in range(X.shape[1])]
            for imp_dict in imp_dicts:
                for i in imp_dict:
                    scores[i] = imp_dict[i]

            os.remove(model_loc)

            if changed is not None:
                model.n_jobs = changed

        scores = np.array(scores)
        difs = baseline_score - scores
        return difs

    def get_feat_importance(self, ind, scorer, model):

        X_copy = self.X.copy()

        importances = []
        for _ in range(self.n_perm):

            X_copy[:, ind] = np.random.permutation(X_copy[:, ind])
            importances.append(scorer(model, X_copy, self.y))

        return np.mean(importances)



'''
Let's store some global-ish pre-set variables in here
'''

SCOPES = set(['float', 'data', 'data files',
              'float covars', 'fc', 'all',
              'n', 'cat', 'categorical', 'covars'])

ORDERED_NAMES = ['loaders', 'imputers', 'scalers',
                 'transformers', 'feat_selectors', 'model']


def is_f2b(d_type):

    valid_names = ['f2b', 'float_to_bin', 'float to bin',
                   'f2c', 'float_to_cat', 'float to cat',
                   'ftb', 'ftc', 'float_to_categorical',
                   'float to categorical', 'float_to_c',
                   'float to c']

    return d_type in valid_names



from ..helpers import to_memmap, from_memmap, get_grid_params, is_ng
import numpy as np
import os
import nevergrad as ng
from nose.tools import raises


def test_to_memmap():

    X = np.ones((10, 10), dtype='float32')
    f, dtype, shape = to_memmap(X)

    assert os.path.exists(f)
    assert dtype == 'float32'
    assert shape == X.shape

    os.unlink(f)
    assert not os.path.exists(f)


def test_from_memmap():

    X = np.ones((10, 10), dtype='float32')
    X_m = to_memmap(X)
    X_c = from_memmap(X_m)

    assert X_c.shape == X.shape
    assert X.dtype == X_c.dtype

    X_c[:] = 0
    X_c = from_memmap(X_m)
    assert np.all(X == 1)


def test_get_grid_params():

    params = {'1': ng.p.Choice([1, 2, 3]),
              '2': ng.p.TransitionChoice([1, 2]),
              '3': 3,
              '4': ng.p.Scalar(lower=1, upper=2).set_integer_casting()}

    grid_params = get_grid_params(params)

    assert grid_params['1'] == [1, 2, 3]
    assert grid_params['2'] == [1, 2]
    assert grid_params['3'] == 3
    assert grid_params['4'] == [1, 2]


@raises(RuntimeError)
def test_get_grid_params_fail():

    params = {'1': ng.p.Scalar(lower=1, upper=2)}
    get_grid_params(params)


def test_is_ng():

    assert not is_ng(6)
    assert not is_ng([1, 2, 3])
    assert not is_ng(set([5]))
    assert not is_ng(from_memmap)
    assert not is_ng(np.ones(10))
    assert is_ng(ng.p.Choice([1, 2]))
    assert is_ng(ng.p.TransitionChoice([1, 2]))
    assert is_ng(ng.p.Scalar())
    assert is_ng(ng.p.Log(init=1))



import numpy as np
from numpy.random import RandomState
from sklearn.base import BaseEstimator


class RandomParcels(BaseEstimator):

    def __init__(self, geo, n_parcels, medial_wall_inds=None,
                 medial_wall_mask=None, random_state=None):
        ''' This object is used to create random surface parcellations

        Parameters
        -----------
        geo : list of array-like
            geo, or geometry, should be passed as a list of arrays
            (or sets, or lists), where
            each index of the outer list is associated with that vertex index.
            Each list should contain all of that vertex's valid nieghbors.
            So for example, if the vertex at index 0 has nieghbors 5, 8, 9, 10,
            then the first element of passed geo should be [5,8,9,10].

        n_parcels : int
            The number of random parcels to generate.
            Will generate as 1-n_parcels, with 0 reserved for
            non-parcels.

        medial_wall_inds : array-like of int or None, optional
            By passing an array-like containing valid ints
            between 0 and the size of the parcel, this parameter
            can specify optionally index where parcels should not be
            generated.

            This parameter should be left as None if
            passing medial_wall_mask, as only one will be used.

            ::

                default = None

        medial_wall_mask : array-like of bool or None, optional
            Simmilar to medial_wall_inds, you may also specify
            the medial wall, or spots not to fill, via this
            parameter by passing a boolean mask, where 1 or True
            signifies that vertex is part of the medial wall to exclude.

            This parameter should be left as None if
            passing medial_wall_inds, as only one will be used.

            ::

                default = None

        random_state : int, None, or RandomState, optional
            A random seed to be passed to create a numpy RandomState,
            as either a numpy random state, an integer seed, or None
            to generate a random RandomState.

            ::

                default = None
        '''

        # Set passed params
        self.geo = geo
        self.n_parcels = n_parcels
        self.medial_wall_inds = medial_wall_inds
        self.medial_wall_mask = medial_wall_mask
        self.random_state = random_state
        self.mask = None

    def get_parc(self, copy=True):

        if self.mask is None:
            self._generate_parc_from_params()

        if copy:
            return self.mask.copy()
        else:
            return self.mask

    def _generate_parc_from_params(self):

        # Proc by input args
        self._proc_medial_wall()
        self._proc_geo()
        self._proc_random_state()

        # Set up mask, done and flags
        self.sz = len(self._geo)
        self.reset()

        # Init
        self.init_parcels()

        # Then generate
        self.generate_parcels()

    def _proc_geo(self):
        '''Convert geo to list of np arrays. Also set so only
        consider neighbors as non medial wall, if any medial wall passed.'''
        self._geo = []
        for g in self.geo:
            valid_n = list(set(g) - self.m_wall)
            self._geo.append(np.array(valid_n))

    def _proc_medial_wall(self):

        # Proc medial wall inds
        if self.medial_wall_inds is not None:
            self.m_wall = set(list(self.medial_wall_inds))
        elif self.medial_wall_mask is not None:
            self.m_wall = set(list(np.where(self.medial_wall_mask)[0]))
        else:
            self.m_wall = set()

    def _proc_random_state(self):

        if self.random_state is None:
            self.r_state = RandomState()
        elif isinstance(self.random_state, int):
            self.r_state = RandomState(seed=self.random_state)
        else:
            self.r_state = self.random_state

    def reset(self):
        '''Just reset the mask, and set w/ done info'''

        self.mask = np.zeros(self.sz, dtype='int16')
        self.done = self.m_wall.copy()
        self.ready, self.generated = False, False

    def init_parcels(self):

        # Generate the starting locs
        valid = np.setdiff1d(np.arange(self.sz), np.array(list(self.done)))
        self.start_locs = self.r_state.choice(valid, size=self.n_parcels,
                                              replace=False)

        # Set random probs. that each loc is chosen
        self.probs = self.r_state.random(size=self.n_parcels)

    def setup(self):
        '''This should be called before generating parcel,
        so after a mutation has been made, setup needs to
        be called. It also does not hurt to call setup an
        extra time, as nothing random is set.'''

        # Generate corresponding labels w/ each loc
        self.labels = np.arange(1, self.n_parcels+1, dtype='int16')

        # Mask where if == 1, then that parcel is done
        self.finished = np.zeros(self.n_parcels, dtype='bool_')

        # Drop the first points
        self.mask[self.start_locs] = self.labels

        # Set ready flag to True
        self.ready = True

    def get_probs(self):

        return self.probs / np.sum(self.probs)

    def choice(self):
        '''Select a valid label based on probs.'''

        msk = self.finished == 0
        probs = self.probs[msk] / np.sum(self.probs[msk])
        label = self.r_state.choice(self.labels[msk], p=probs)

        return label

    def get_valid_neighbors(self, loc):

        ns = self._geo[loc]
        valid_ns = ns[self.mask[ns] == 0]

        return valid_ns

    def generate_parcels(self):

        if self.ready is False:
            self.setup()

        # Keep looping until every spot is filled
        while (self.finished == 0).any():
            self.add_spot()

        # Set generated flag when done
        self.generated = True

    def add_spot(self):

        # Select which parcel to add to
        label = self.choice()

        # Determine valid starting locations anywhere in exisitng parcel
        current = np.where(self.mask == label)[0]
        valid = set(current) - self.done

        self.proc_spot(valid, label)

    def proc_spot(self, valid, label):

        # If no valid choices, then set this parcel to finished
        if len(valid) == 0:
            self.finished[label-1] = 1
            return

        # Select randomly from the valid starting locs
        loc = self.r_state.choice(tuple(valid))

        # Select a valid + open neighbor
        valid_ns = self.get_valid_neighbors(loc)

        if len(valid_ns) > 0:

            # Select a valid choice, and add it w/ the right label
            choice = self.r_state.choice(valid_ns)
            self.mask[choice] = label

            # If this was the only choice, mark start loc as done
            if len(valid_ns) == 1:
                self.done.add(loc)

        # If there are no valid choices, mark as done
        else:
            self.done.add(loc)

            valid.remove(loc)
            self.proc_spot(valid, label)



from .main.BPt_ML import BPt_ML, Load
from .main.BPt_ML import BPt_ML as ABCD_ML
from .main.Params_Classes import (Loader, Imputer, Scaler, Transformer,
                                  Feat_Selector, Model, Ensemble,
                                  Param_Search, Feat_Importance,
                                  Model_Pipeline,
                                  Problem_Spec, Shap_Params,
                                  CV, CV_Split, CV_Splits)

from .main.Input_Tools import (Select, Duplicate, Pipe, Value_Subset,
                               Values_Subset)

__author__ = "sahahn"
__version__ = "1.3.6"
__all__ = ["BPt_ML", "ABCD_ML", "Load", "Loader",
           "Imputer", "Scaler", "Transformer",
           "Feat_Selector", "Model",
           "Ensemble", "Param_Search", "Feat_Importance",
           "Model_Pipeline", "Problem_Spec", "Select",
           "Duplicate", "Pipe", "Value_Subset", "Values_Subset",
           "Shap_Params", "CV", "CV_Split", "CV_Splits"]



"""
main.py
====================================
The main project class.
"""
import pandas as pd
import shutil
import os
import pickle as pkl

from ..helpers.Docstring_Helpers import get_new_docstring
# from ..helpers.Params_Classes import ML_Params
from ..helpers.CV import CV


def Load(loc, exp_name='default', log_dr='default', existing_log='default',
         verbose='default', notebook='default', random_state='default'):
    '''
    This function is designed to load in a saved previously created
    BPt_ML object.

    See :func:`Save <BPt.BPt_ML.Save>` for saving an object.
    See :func:`Init <BPt.BPt_ML>` for the
    rest of changable param descriptions, e.g., log_dr, existing_log, ect...

    Parameters
    ----------
    loc : str or Path

        A path/str to a saved BPt_ML object,
        (One saved with :func:`Save <BPt.BPt_ML.Save>`), then that object will be
        loaded. Notably, if any additional params are passed along
        with it, e.g., exp_name, notebook, ect... they will override
        the saved values with the newly passed values.
        If left as 'default', all params will be set to the loaded value,
        though see the warning below.

        .. WARNING::
            The exp_name or log_dr may need to be changed, especially
            in the case where the object is being loaded in a new
            location or enviroment from where the original was created,
            as it will by default try to create logs with the saved path
            information as the original.

        You can only change exp_name, log_dr, existing_log, verbose,
        notebook and random_state when loading a new object, for the
        remaining params, even if a value is passed, it will not be
        applied. If the user really wishes to change one of these params,
        they can change it manually via self.name_of_param = whatever.
    '''

    with open(loc, 'rb') as f:
        ML = pkl.load(f)

    if exp_name != 'default':
        ML.exp_name = exp_name
    if log_dr != 'default':
        ML.log_dr = log_dr
    if existing_log != 'default':
        ML.existing_log = existing_log
    if verbose != 'default':
        ML.verbose = verbose

    ML._init_logs()

    if notebook != 'default':
        ML.notebook = notebook
    if random_state != 'default':
        ML.random_state = random_state

    ML._print('ML object loaded from save!')
    return ML


class BPt_ML():

    def __init__(self, exp_name='My_Exp', log_dr='', existing_log='append',
                 verbose=True, notebook=True,
                 use_abcd_subject_ids=False,
                 low_memory_mode=False, strat_u_name='_Strat',
                 random_state=534, n_jobs=1, dpi=100, mp_context='loky'):
        '''Main class used within BPt for interfacing with Data Loading
        and Modeling / Other funcationality.

        Parameters
        ----------
        exp_name : str, optional
            The name of this experimental run,
            used explicitly in saving logs, and figures, where the passed
            `exp_name` is used as the name of the log folder.
            If log_dr is not set to None,
            (if not None then saves logs and figures)
            then a folder is created within the log dr
            with the exp_name.

            ::

                default = 'My_Exp'

        log_dr : str, Path or None, optional
            The directory in which to store logs...
            If set to None, then will not save any logs!
            If set to empty str, will save in the current dr.

            ::

                default = ''

        existing_log : {'new', 'append', 'overwrite'}, optional
            This parameter dictates different choices for when
            an a folder with exp_name already exists in the specified
            log_dr.

            These choices are:

            - 'new'
                If the log folder already exists, then
                just increment `exp_name` until a free name is found,
                and use that as the log folder / `exp_name`.

            - 'append'
                If existing_log is 'append' then log entries
                and new figures will be added to the existing folder.

            - 'overwrite'
                If existing_log is 'overwrite', then the existing
                log folder with the same exp_name will be cleared
                upon __init__.

            ::

                default = 'append'

        verbose: bool, optional
            If `verbose` is set to True, the BPt_ML object
            will print output, diagnostic and more general, directly
            to std out. If set to False, no output will be printed, though
            output will still be recorded within the logs assuming
            log_dr is not None.

            ::

                default = True

        notebook : bool, optional
            If True, then assumes the user is running
            the code in an interactive jupyter notebook. 
            In this case, certain features will either be enabled or disabled,
            e.g., type of progress bar.

            ::

                default = Trues

        use_abcd_subject_ids : bool, optional
            Flag to determine the usage of ABCD speficic 'default'
            subject id behavior.
            If set to True, this will convert input NDAR subject ids
            into upper case, with prepended NDAR - type format.
            If set to False, then all input subject names must be entered
            explicitly the same, no preprocessing will be done on them.

            ::

                default = False

        low_memory_mode : bool, optional
            This parameter dictates behavior around loading in data,
            specifically,
            If set to True, individual dataframes self.data, self.covars ect...
            will be deleted from memory as soon as modeling begins.
            This parameter also controls the pandas read_csv behavior,
            which also has a low_memory flag.

            ::

                default = False

        strat_u_name : str, optional
            A unique str identifier to be appended to every loaded
            strat value (to keep them seperate from covars and data).

            You should only need to change or ever worry about this in
            the case that one of your input variables happens to have the
            default value of '_Strat' in it...

            ::

                default = '_Strat'

        random_state : int, RandomState instance or None, optional
            The default random state, either as int for a specific seed,
            or if None then the random seed is set by np.random.
            This parameters if set will be the default random_state class-wide,
            so any place random_state is left to default, unless a different
            default is set (e.g. default load value or default ML value) this
            random state will be used.

            ::

                default = 534

        n_jobs : int, optional
            The default number of jobs / processors to use (if avaliable) where
            ever avaliable class-wide across the BPt.

            ::

                default = 1

        dpi : int, optional
            The default dpi in which to save any automatically saved fiugres
            with.
            Where this parameter can also be set to specific values
            for specific plots.

            ::

                default = 1

        mp_context : str, optional

            When a hyper-parameter search is launched, there are different
            ways through python that the multi-processing can be launched
            (assuming n_jobs > 1). Occassionally some choices can lead to
            unexpected errors.

            Choices are:

            - 'loky': Create and use the python library
                loky backend.

            - 'fork': Python default fork mp_context

            - 'forkserver': Python default forkserver mp_context

            - 'spawn': Python default spawn mp_context

            ::

                default = 'loky'
        '''
        # Load logging class params
        self.exp_name = exp_name
        self.log_dr = log_dr
        self.existing_log = existing_log
        self.verbose = verbose

        self._init_logs()

        self._print('exp_name =', self.exp_name)
        self._print('log_dr =', self.log_dr)
        self._print('existing_log =', self.existing_log)
        self._print('verbose =', self.verbose)
        self._print('exp log dr setup at:', self.exp_log_dr)
        self._print('log file at:', self.log_file)

        # Set rest of class params
        self.notebook = notebook
        self.use_abcd_subject_ids = use_abcd_subject_ids
        self.low_memory_mode = low_memory_mode
        self.strat_u_name = strat_u_name
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.dpi = dpi
        self.mp_context = mp_context

        self._print('Default params set:')
        self._print('notebook =', self.notebook)
        self._print('use_abcd_subject_ids =', self.use_abcd_subject_ids)
        self._print('low memory mode =', self.low_memory_mode)
        self._print('strat_u_name =', self.strat_u_name)
        self._print('random state =', self.random_state)
        self._print('n_jobs =', self.n_jobs)
        self._print('dpi =', self.dpi)
        self._print('mp_context =', self.mp_context)

        # Initialze various variables
        self.name_map, self.exclusions, self.inclusions = {}, set(), set()
        self.data, self.covars = pd.DataFrame(), pd.DataFrame()
        self.targets, self.strat = pd.DataFrame(), pd.DataFrame()

        # Dict objects to hold encoders
        self.covars_encoders = {}
        self.targets_encoders = {}
        self.strat_encoders = {}

        # Class values to be set later
        self.all_data = None
        self.targets_keys = []

        # Stores the gloabl train/test split
        self.train_subjects, self.test_subjects = None, None

        # CV by default is just random splits
        self.cv = CV()

        # Store default dicts as init empty
        self.default_load_params, self.default_ML_verbosity = {}, {}

        self.subject_id = 'src_subject_id'

        self.file_mapping = {}
        self.data_file_keys = []

        self._print('BPt_ML object initialized')

    def Save(self, loc, low_memory=False):
        '''This class method is used to save an existing BPt_ML
        object for further use.

        Parameters
        ----------
        loc : str or Path
            The location in which the pickle of the BPt_ML object
            should be saved! This is the same loc which should be
            passed to :func:`Load <BPt.main.BPt_ML.Load>` in order to
            re-load the object.

        low_memory : bool, optional
            If this parameter is set to True, then self.data,
            self.targets, self.covars, self.strat will be deleted
            before saving. The assumption for the param to be used is
            that self.all_data has already been created, and therefore
            the individual dataframes with data, covars ect... can safely
            be deleted as the user will not need to work with them directly
            any more.

            ::

                default = False
        '''

        if low_memory:
            self.data, self.covars = pd.DataFrame(), pd.DataFrame()
            self.targets, self.strat = pd.DataFrame(), pd.DataFrame()

            try:
                del self.evaluator
            except AttributeError:
                pass

        with open(loc, 'wb') as f:
            pkl.dump(self, f)

    def _init_logs(self):

        if self.log_dr is not None:

            if self.log_dr == '':
                self.log_dr = os.getcwd()

            # Ensure log_dr exists, if not make it
            os.makedirs(self.log_dr, exist_ok=True)

            # Get exp_log_dr name
            self.exp_log_dr = os.path.join(self.log_dr, self.exp_name)

            if os.path.isdir(self.exp_log_dr):

                if self.existing_log == 'new':

                    cnt = 1
                    while os.path.isdir(self.exp_log_dr +
                                        '(' + str(cnt) + ')'):
                        cnt += 1

                    self.exp_log_dr += '(' + str(cnt) + ')'

                # If overwrite, delete everything, then make new blank
                elif self.existing_log == 'overwrite':
                    shutil.rmtree(self.exp_log_dr)

                # Make the new dr
                if self.existing_log != 'append':
                    os.mkdir(self.exp_log_dr)

            # If the dr doesn't already exist, regardless of existing log
            # Just make new dr.
            else:
                os.mkdir(self.exp_log_dr)

            # Make the log file if not already made.
            self.log_file = os.path.join(self.exp_log_dr, 'logs.txt')

        else:
            self.exp_log_dr = None
            self.log_file = None

    def _print(self, *args, **kwargs):
        '''Overriding the print function to allow for
        customizable verbosity within class methods. Will also
        take care of logging behavior.

        Parameters
        ----------
        args
            Anything that would be passed to default python print
        '''

        dont_print = kwargs.pop('dont_print', False)

        if self.verbose and not dont_print:
            print(*args, **kwargs)

        if self.log_file is not None:
            log = open(self.log_file, 'a')
            print(*args, **kwargs, file=log)
            log.close()

    def _print_nothing(self, *args, **kwargs):
        pass

    # Data loader functionality
    from ._Data import (Set_Default_Load_Params,
                        _make_load_params,
                        _get_data_file_cnt,
                        Load_Name_Map,
                        Load_Data,
                        Load_Data_Files,
                        Load_Targets,
                        _proc_target,
                        _print_loaded_targets,
                        Load_Covars,
                        _proc_covar,
                        Load_Strat,
                        _proc_strat,
                        Load_Exclusions,
                        Load_Inclusions,
                        Drop_Data_Cols,
                        _drop_data_cols,
                        Filter_Data_Cols,
                        Filter_Data_Files_Cols,
                        Consolidate_Data_Files,
                        Proc_Data_Unique_Cols,
                        _proc_data_unique_cols,
                        Drop_Data_Duplicates,
                        Binarize_Target,
                        _proc_threshold,
                        Binarize_Covar,
                        Get_Overlapping_Subjects,
                        Clear_Name_Map,
                        Clear_Data,
                        Clear_Covars,
                        Clear_Targets,
                        Clear_Strat,
                        Clear_Exclusions,
                        Clear_Inclusions,
                        Get_Nan_Subjects,
                        _get_targets_key,
                        _load_datasets,
                        _load_user_passed,
                        _load_dataset,
                        _common_load,
                        _load,
                        _set_overlap,
                        _merge_existing,
                        _proc_df,
                        _load_set_of_subjects,
                        _process_subject_name,
                        _drop_na,
                        _filter_by_eventname,
                        _show_na_info,
                        _drop_excluded,
                        _drop_included,
                        _filter_excluded,
                        _filter_included,
                        _get_overlapping_subjects,
                        Prepare_All_Data,
                        _get_cat_keys,
                        _set_data_scopes,
                        _get_base_targets_names)

    # Update loader docstrings
    Load_Name_Map.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Load_Name_Map)
    Load_Data.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Load_Data)
    Load_Data_Files.__doc__ =\
        get_new_docstring(Load_Data, Load_Data_Files)
    Load_Targets.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Load_Targets)
    Load_Covars.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Load_Covars)
    Load_Strat.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Load_Strat)
    Filter_Data_Cols.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Filter_Data_Cols)
    Proc_Data_Unique_Cols.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Proc_Data_Unique_Cols)
    Drop_Data_Duplicates.__doc__ =\
        get_new_docstring(Set_Default_Load_Params, Drop_Data_Duplicates)

    # Validation / CV funcationality
    from ._Validation import (_get_cv,
                              Define_Validation_Strategy,
                              Train_Test_Split,
                              _add_strat_u_name,
                              _get_info_on)

    # Machine Learning functionality
    from ._ML import (Set_Default_ML_Verbosity,
                      _ML_print,
                      Evaluate,
                      Test,
                      _premodel_check,
                      _preproc_param_search,
                      _preproc_model_pipeline,
                      _preproc_cv_splits,
                      _preproc_problem_spec,
                      _get_split_vals,
                      _get_subjects_to_use,
                      _init_evaluator,
                      _handle_scores,
                      _print_summary_score,
                      _save_results,
                      get_pipeline)

    from ._Plotting import (_plot,
                            _proc_subjects,
                            _get_plot_df,
                            Show_Data_Dist,
                            _input_targets,
                            _input_covars,
                            _input_strat,
                            Show_Targets_Dist,
                            Show_Covars_Dist,
                            Show_Strat_Dist,
                            _get_single_df,
                            _show_single_dist,
                            _get_cat_display_df,
                            _show_dist,
                            _display_df,
                            _get_top_global,
                            Plot_Global_Feat_Importances,
                            _plot_multiclass_global_feat_importances,
                            _plot_global_feat_importances,
                            Plot_Local_Feat_Importances,
                            _plot_shap_summary)

    from ._Tables import (Save_Table,
                          _get_single_dfs,
                          _get_table_contents,
                          _get_group_titles)
